import warnings
from functools import cached_property
from math import isclose
from typing import Sequence, Literal

import astropy.units as u
import numpy as np
from astropy.modeling import models, Model, fitting, CompoundModel
from matplotlib.pyplot import Axes, Figure, setp, subplots
from numpy.ma import MaskedArray
from numpy.typing import ArrayLike
from scipy import optimize, ndimage
from scipy.spatial import KDTree

from specreduce.calibration_data import load_pypeit_calibration_lines
from specreduce.compat import Spectrum
from specreduce.line_matching import find_arc_lines

__all__ = ["WavelengthCalibration1D"]

from specreduce.wavesol1d import WavelengthSolution1D


def _format_linelist(lst: ArrayLike) -> MaskedArray:
    """Force a line list into a MaskedArray with a shape of (n, 2) where n is the number of lines.

    Parameters
    ----------
    lst
        Input array of centroids or centroids with amplitudes. Must be either:
            - A 1D array with a shape [n] for centroids.
            - A 2D array with a shape [n, 2] for centroids and amplitudes.

    Returns
    -------
    numpy.ma.MaskedArray
        Formatted and standardized line list array with shape [n, 2], where each row
        contains a line centroid and amplitude.

    Raises
    ------
    ValueError
        If the input line list does not meet the specified dimensional or shape
        requirements.
    """
    lst: MaskedArray = MaskedArray(lst, copy=True)
    lst.mask = np.ma.getmaskarray(lst)

    if lst.ndim > 2 or lst.ndim == 2 and lst.shape[1] > 2:
        raise ValueError(
            "Line lists must be 1D with a shape [n] (centroids) or "
            "2D with a shape [n, 2] (centroids and amplitudes)."
        )

    if lst.ndim == 1:
        lst = MaskedArray(np.tile(lst[:, None], [1, 2]))
        lst[:, 1] = 0.0
        lst.mask[:, :] = lst.mask.any(axis=1)[:, None]

    return lst[np.argsort(lst.data[:, 0])]


def _unclutter_text_boxes(labels: Sequence) -> None:
    """Remove overlapping labels from the plot.

    Removes overlapping text labels from a set of matplotlib label objects. The function iterates
    over all combinations of labels, checks for overlaps among their bounding boxes, and removes
    the label with the lower z-order in case of an overlap.

    Parameters
    ----------
    labels
        A list of matplotlib.text.Text objects.
    """
    to_remove = set()
    for i in range(len(labels)):
        for j in range(i + 1, len(labels)):
            l1 = labels[i]
            l2 = labels[j]
            bbox1 = l1.get_window_extent()
            bbox2 = l2.get_window_extent()
            if bbox1.overlaps(bbox2):
                if l1.zorder < l2.zorder:
                    to_remove.add(l1)
                else:
                    to_remove.add(l2)

    for label in to_remove:
        label.remove()


class WavelengthCalibration1D:
    def __init__(
        self,
        arc_spectra: Spectrum | Sequence[Spectrum] | None = None,
        obs_lines: ArrayLike | Sequence[ArrayLike] | None = None,
        line_lists: ArrayLike | None = None,
        unit: u.Unit = u.angstrom,
        ref_pixel: float | None = None,
        pix_bounds: tuple[int, int] | None = None,
        line_list_bounds: None | tuple[float, float] = None,
        n_strogest_lines: None | int = None,
        wave_air: bool = False,
    ) -> None:
        """A class for wavelength calibration of one-dimensional spectra.

        This class is designed to facilitate wavelength calibration of one-dimensional spectra,
        with support for both direct input of line lists and observed spectra. It uses a polynomial
        model for fitting the wavelength solution and offers features to incorporate catalog lines
        and observed line positions.

        Parameters
        ----------
        arc_spectra
            Arc spectra provided as ``Spectrum`` objects for wavelength fitting, by default
            None. This parameter and ``obs_lines`` cannot be provided simultaneously.

        obs_lines
            Pixel positions of observed spectral lines for wavelength fitting, by default None. This
            parameter and ``arc_spectra`` cannot be provided simultaneously.

        line_lists
            Catalogs of spectral line wavelengths for wavelength calibration. Provide either an
            array of line wavelengths or a list of `PypeIt <https://github.com/pypeit/PypeIt>`_
            catalog names. If `None`, no line lists are used. You can query the list of available
            catalog names via `~specreduce.calibration_data.get_available_line_catalogs`.

        unit
            The unit of the wavelength calibration, by default ``astropy.units.Angstrom``.

        ref_pixel
            The reference pixel in which the wavelength solution will be centered.

        pix_bounds
            Lower and upper pixel bounds for fitting, defined as a tuple (min, max). If
            ``obs_lines`` is provided, this parameter is mandatory.

        line_list_bounds
            Wavelength bounds as a tuple (min, max) for filtering usable spectral
            lines from the provided line lists.

        n_strogest_lines
            The number of strongest lines to be included from the line lists. If `None`, all
            are included.

        wave_air
            Boolean indicating whether the input wavelengths correspond to air rather than vacuum;
            by default `False`, meaning vacuum wavelengths.
        """
        self.unit = unit
        self._unit_str = unit.to_string("latex")
        self.degree = None
        self.ref_pixel = ref_pixel
        self.nframes = 0

        if ref_pixel is not None and ref_pixel < 0:
            raise ValueError("Reference pixel must be positive.")

        self.arc_spectra: list[Spectrum] | None = None
        self.bounds_pix: tuple[int, int] | None = pix_bounds
        self.bounds_wav: tuple[float, float] | None = None
        self._cat_lines: list[MaskedArray] | None = None
        self._obs_lines: list[MaskedArray] | None = None
        self._trees: list[KDTree] | None = None

        self._fit: optimize.OptimizeResult | None = None
        self.solution = WavelengthSolution1D(None, pix_bounds, unit)

        # Read and store the observational data if given. The user can provide either a list of arc
        # spectra as Spectrum objects or a list of line pixel position arrays. An attempt to give
        # both raises an error.
        if arc_spectra is not None and obs_lines is not None:
            raise ValueError("Only one of arc_spectra or obs_lines can be provided.")

        if arc_spectra is not None:
            self.arc_spectra = [arc_spectra] if isinstance(arc_spectra, Spectrum) else arc_spectra
            self.nframes = len(self.arc_spectra)
            for s in self.arc_spectra:
                if s.data.ndim > 1:
                    raise ValueError("The arc spectrum must be one dimensional.")
            if len(set([s.data.size for s in self.arc_spectra])) != 1:
                raise ValueError("All arc spectra must have the same length.")
            self.bounds_pix = (0, self.arc_spectra[0].shape[0])
            self.solution.bounds_pix = self.bounds_pix
            if self.ref_pixel is None:
                self.ref_pixel = self.arc_spectra[0].data.size / 2

        elif obs_lines is not None:
            self.observed_lines = obs_lines
            self.nframes = len(self._obs_lines)
            if self.bounds_pix is None:
                raise ValueError("Must give pixel bounds when providing observed line positions.")
            if self.ref_pixel is None:
                raise ValueError("Must give reference pixel when providing observed lines.")

        # Read the line lists if given. The user can provide an array of line wavelength positions
        # or a list of line list names (used by `load_pypeit_calibration_lines`) for each arc
        # spectrum.
        if line_lists is not None:
            if not isinstance(line_lists, (tuple, list)):
                line_lists = [line_lists]

            if len(line_lists) != self.nframes:
                raise ValueError("The number of line lists must match the number of arc spectra.")
            self._read_linelists(
                line_lists,
                line_list_bounds=line_list_bounds,
                wave_air=wave_air,
                n_strongest=n_strogest_lines,
            )

    def _line_match_distance(self, x: ArrayLike, model: Model, max_distance: float = 100) -> float:
        """Compute the sum of distances between catalog lines and transformed observed lines.

        This function evaluates the pixel-to-wavelength model at the observed line positions,
        queries the nearest catalog line (via KDTree), and sums the distances after clipping
        them at `max_distance`. The result is suitable as a scalar objective for global
        optimization of the wavelength solution.

        Parameters
        ----------
        x
            Pixel-to-wavelength model parameters (e.g., Polynomial1D coefficients c0..cN).
        model
            The pixel-to-wavelength model to be evaluated.
        max_distance
            Upper bound used to clip individual distances before summation.

        Returns
        -------
        float
            Sum of nearest-neighbor distances between transformed observed lines and catalog lines.

        """
        total_distance = 0.0
        for t, l in zip(self._trees, self.observed_line_locations):
            transformed_lines = model.evaluate(l, -self.ref_pixel, *x)[:, None]
            total_distance += np.clip(t.query(transformed_lines)[0], 0, max_distance).sum()
        return total_distance

    def _read_linelists(
        self,
        line_lists: Sequence,
        line_list_bounds: None | tuple[float, float] = None,
        wave_air: bool = False,
        n_strongest: None | int = None,
    ) -> None:
        """Read and processes line lists.

        Parameters
        ----------
        line_lists
            A collection of line lists that can either be arrays of wavelengths or `PypeIt
            <https://github.com/pypeit/PypeIt>`_
            lamp names. You can query the list of available catalog names via
            `~specreduce.calibration_data.get_available_line_catalogs`.

        line_list_bounds
            A tuple specifying the minimum and maximum wavelength bounds. Only wavelengths
            within this range are retained.

        wave_air
             If True, convert the vacuum wavelengths used by `PypeIt
             <https://github.com/pypeit/PypeIt>`_ to air wavelengths.

        n_strongest
            The number of strongest lines to be used. If `None`, all lines are used.
        """

        lines = []
        for lst in line_lists:
            if isinstance(lst, np.ndarray):
                lines.append(lst)
            else:
                if isinstance(lst, str):
                    lst = [lst]
                lines.append([])
                for ll in lst:
                    line_table = load_pypeit_calibration_lines(ll, wave_air=wave_air)
                    if n_strongest is not None:
                        ix = np.argsort(line_table["amplitude"].value)[::-1]
                        lines[-1].append(line_table[ix][:n_strongest]["wavelength"].to(
                            self.unit).value)
                    else:
                        lines[-1].append(line_table["wavelength"].to(self.unit).value)
                lines[-1] = np.ma.masked_array(np.sort(np.concatenate(lines[-1])))

        if line_list_bounds is not None:
            for i, lst in enumerate(lines):
                lines[i] = lst[(lst >= line_list_bounds[0]) & (lst <= line_list_bounds[1])]

        self.catalog_lines = lines
        self._create_trees()

    def _create_trees(self) -> None:
        """Initialize the KDTree instances for the current set of catalog line locations."""
        self._trees = [KDTree(lst.compressed()[:, None]) for lst in self.catalog_line_locations]

    def find_lines(self, fwhm: float, noise_factor: float = 1.0) -> None:
        """Find lines in the provided arc spectra.

        Determines the spectral lines within each spectrum of the arc spectra based on the
        provided initial guess for the line Full Width at Half Maximum (FWHM).

        Parameters
        ----------
        fwhm
            Initial guess for the FWHM for the spectral lines, used as a parameter in
            the ``find_arc_lines`` function to locate and identify spectral arc lines.

        noise_factor
            The factor to multiply the uncertainty by to determine the noise threshold
            in the `~specutils.fitting.find_lines_threshold` routine.
        """
        if self.arc_spectra is None:
            raise ValueError("Must provide arc spectra to find lines.")

        line_lists = []
        for i, arc in enumerate(self.arc_spectra):
            lines = find_arc_lines(arc, fwhm, noise_factor=noise_factor)
            ix = np.round(lines["centroid"].value).astype(int)
            if np.any((ix < 0) | (ix >= arc.shape[0])):
                raise ValueError(
                    "Error in arc line identification. Try increasing ``noise_factor``."
                )
            amplitudes = ndimage.maximum_filter1d(arc.flux.value, 5)[ix]
            line_lists.append(
                np.ma.masked_array(np.transpose([lines["centroid"].value, amplitudes]))
            )
        self.observed_lines = line_lists

    def _create_model(self, degree: int, coeffs: None | ArrayLike = None) -> CompoundModel:
        """Initialize the polynomial model with the given degree and an optional base model.

        This method sets up a polynomial transformation based on the reference pixel and degree.
        If coefficients are provided, they are copied to the initialized model up to the degree
        specified.

        Parameters
        ----------
        degree
            Degree of the polynomial model to be initialized.

        coeffs
            Optional initial polynomial coefficients.
        """
        self.degree = degree

        pars = {}
        if coeffs is not None:
            nc = min(degree + 1, len(coeffs))
            pars = {f"c{i}": c for i, c in enumerate(coeffs[:nc])}
        return models.Shift(-self.ref_pixel) | models.Polynomial1D(self.degree, **pars)

    def fit_lines(
        self,
        pixels: ArrayLike,
        wavelengths: ArrayLike,
        degree: int = 3,
        match_obs: bool = False,
        match_cat: bool = False,
        refine_fit: bool = True,
        refine_max_distance: float = 5.0,
        refined_fit_degree: int | None = None,
    ) -> WavelengthSolution1D:
        """Fit the pixel-to-wavelength model using provided line pairs.

        This method fits the pixel-to-wavelength transformation using explicitly provided pairs
        of pixel coordinates and their corresponding wavelengths via a linear least-squares fit

        Optionally, the provided pixel and wavelength values can be "snapped" to the nearest
        values present in the internally stored observed line list and catalog line list,
        respectively.  This allows the inputs to be approximate, as the snapping step selects
        the nearest precise centroids and catalog values when available.

        Parameters
        ----------
        pixels
            An array of pixel positions corresponding to known spectral lines.

        wavelengths
            An array of the same size as ``pixels``, containing the known
            wavelengths corresponding to the given pixel positions.

        degree
            The polynomial degree for the wavelength solution.

        match_obs
            If True, snap the input ``pixels`` values to the nearest
            pixel values found in ``self.observed_line_locations`` (if available). This helps
            ensure the fit uses the precise centroids detected by `find_lines`
            or provided initially.

        match_cat
            If True, snap the input ``wavelengths`` values to the
            nearest wavelength values found in ``self.catalog_line_locations`` (if available).
            This ensures the fit uses the precise catalog wavelengths.

        refine_fit
            If True (default), automatically call the ``refine_fit`` method
            immediately after the global optimization to improve the solution
            using a least-squares fit on matched lines.

        refine_max_distance
            Maximum allowed separation between catalog and observed lines for them to
            be considered a match during ``refine_fit``. Ignored if ``refine_fit`` is False.

        refined_fit_degree
            The polynomial degree for the refined fit. Can be higher than ``degree``. If ``None``,
            equals to ``degree``.
        """
        pixels = np.asarray(pixels)
        wavelengths = np.asarray(wavelengths)
        if pixels.size != wavelengths.size:
            raise ValueError("The sizes of pixel and wavelength arrays must match.")
        nlines = pixels.size
        if nlines < 2:
            raise ValueError("Need at least two lines for a fit")

        if self.bounds_pix is None:
            raise ValueError("Cannot fit without pixel bounds set.")

        # Match the input wavelengths to catalog lines.
        if match_cat:
            if self._cat_lines is None:
                raise ValueError("Cannot fit without catalog lines set.")
            tree = KDTree(
                np.concatenate([c.compressed() for c in self.catalog_line_locations])[:, None]
            )
            ix = tree.query(wavelengths[:, None])[1]
            wavelengths = tree.data[ix][:, 0]

        # Match the input pixel values to observed pixel values.
        if match_obs:
            if self._obs_lines is None:
                raise ValueError("Cannot fit without observed lines set.")
            tree = KDTree(
                np.concatenate([c.compressed() for c in self.observed_line_locations])[:, None]
            )
            ix = tree.query(pixels[:, None])[1]
            pixels = tree.data[ix][:, 0]

        fitter = fitting.LinearLSQFitter()
        shift, model = self._create_model(degree)
        if model.degree > nlines:
            warnings.warn(
                "The degree of the polynomial model is higher than the number of lines. "
                "Fixing the higher-order coefficients to zero."
            )
            for i in range(nlines, model.degree + 1):
                model.fixed[f"c{i}"] = True
        model = fitter(model, pixels - self.ref_pixel, wavelengths)
        for i in range(model.degree + 1):
            model.fixed[f"c{i}"] = False
        self.solution.p2w = shift | model

        can_match = self._cat_lines is not None and self._obs_lines is not None
        if refine_fit and can_match:
            self.refine_fit(refined_fit_degree, max_match_distance=refine_max_distance)
        else:
            if can_match:
                self.match_lines()
        return self.solution

    def fit_dispersion(
        self,
        wavelength_bounds: tuple[float, float],
        dispersion_bounds: tuple[float, float],
        higher_order_limits: Sequence[float] | None = None,
        degree: int = 3,
        popsize: int = 30,
        max_distance: float = 100,
        refine_fit: bool = True,
        refine_max_distance: float = 5.0,
        refined_fit_degree: int | None = None,
    ) -> WavelengthSolution1D:
        """Calculate a wavelength solution using all the catalog and observed lines.

        This method estimates a wavelength solution without pre-matched pixel–wavelength
        pairs, making it suitable for automated pipelines on stable, well-characterized
        spectrographs. It uses differential evolution to optimize the polynomial parameters
        that minimize the distance between the predicted wavelengths of the observed lines
        and their nearest catalog lines. The resulting solution can optionally be refined
        with a least-squares fit to automatically matched lines.

        Parameters
        ----------
        wavelength_bounds
            (min, max) bounds for the wavelength at ``ref_pixel``; used as an optimization
            constraint.

        dispersion_bounds
            (min, max) bounds for the dispersion d(wavelength)/d(pixel) at ``ref_pixel``; used
            as an optimization constraint.

        higher_order_limits
            Absolute limits for the higher-order polynomial coefficients. Each coefficient is
            constrained to [-limit, limit]. If provided, the number of limits must equal
            (polynomial degree - 1).

        degree
            The polynomial degree for the wavelength solution.

        popsize
            Population size for ``scipy.optimize.differential_evolution``. Larger values can
            improve the chance of finding the global minimum at the cost of additional time.

        max_distance
            Maximum wavelength separation used when associating observed and catalog lines in
            the optimization. Distances larger than this threshold are clipped to this value
            in the cost function to limit the impact of outliers.

        refine_fit
            If True (default), call ``refine_fit`` after global optimization to improve the
            solution using a least-squares fit on matched lines.

        refine_max_distance
            Maximum allowed separation between catalog and observed lines for them to
            be considered a match during ``refine_fit``. Ignored if ``refine_fit`` is False.

        refined_fit_degree
            The polynomial degree for the refined fit. Can be higher than ``degree``. If ``None``,
            equals to ``degree``.
        """

        # Define bounds for differential_evolution.
        bounds = [np.asarray(wavelength_bounds), np.asarray(dispersion_bounds)]
        model = self._create_model(degree)

        if higher_order_limits is not None:
            if len(higher_order_limits) != model[1].degree - 1:
                raise ValueError(
                    "The number of higher-order limits must match the degree of the polynomial "
                    "model minus one."
                )
            for v in higher_order_limits:
                bounds.append(np.asarray([-v, v]))
        else:
            for i in range(2, model[1].degree + 1):
                bounds.append(
                    np.array([-1, 1]) * 10 ** (np.log10(np.mean(dispersion_bounds)) - 2 * i)
                )
        bounds = np.array(bounds)

        self._fit = optimize.differential_evolution(
            lambda x: self._line_match_distance(x, model, max_distance),
            bounds=bounds,
            popsize=popsize,
            init="sobol",
        )
        self.solution.p2w = self._create_model(degree, coeffs=self._fit.x)

        can_match = self._cat_lines is not None and self._obs_lines is not None
        if refine_fit:
            self.refine_fit(refined_fit_degree, max_match_distance=refine_max_distance)
        else:
            if can_match:
                self.match_lines()
        return self.solution

    def refine_fit(
        self, degree: None | int = None, max_match_distance: float = 5.0, max_iter: int = 5
    ) -> WavelengthSolution1D:
        """Refine the pixel-to-wavelength transformation fit.

        Fits (or re-fits) the polynomial wavelength solution using the currently
        matched pixel–wavelength pairs. Optionally adjusts the polynomial degree,
        filters matches by a maximum pixel-space separation, and iterates the fit.

        Parameters
        ----------
        degree
            The polynomial degree for the wavelength solution. If ``None``, the degree
            previously set by the `~WavelengthCalibration1D.fit_lines` or
            `~WavelengthCalibration1D.fit_dispersion` method will be used.

        max_match_distance
            Maximum allowable distance used to identify matched pixel and wavelength
            data points. Points exceeding the bound will not be considered in the fit.

        max_iter
            Maximum number of fitting iterations.
        """

        # Create a new model with the current parameters if degree is specified.
        if degree is not None and degree != self.degree:
            model = self._create_model(degree, coeffs=self.solution.p2w[1].parameters)
        else:
            model = self.solution.p2w

        shift, poly = model
        fitter = fitting.LinearLSQFitter()
        rms = np.nan
        for i in range(max_iter):
            self.match_lines(max_match_distance)
            matched_pix = np.ma.concatenate(self.observed_line_locations).compressed()
            matched_wav = np.ma.concatenate(self.catalog_line_locations).compressed()
            rms_new = np.sqrt(((matched_wav - model(matched_pix)) ** 2).mean())
            if isclose(rms_new, rms):
                break
            model = shift | fitter(poly, matched_pix - self.ref_pixel, matched_wav)
            rms = rms_new
        self.solution.p2w = model
        return self.solution

    @property
    def degree(self) -> None | int:
        return self._degree

    @degree.setter
    def degree(self, degree: int | None):
        if degree is not None and degree < 1:
            raise ValueError("Degree must be at least 1.")
        self._degree = degree

    @property
    def observed_lines(self) -> None | list[MaskedArray]:
        """Pixel positions and amplitudes of the observed lines as a list of masked arrays."""
        return self._obs_lines

    @cached_property
    def observed_line_locations(self) -> None | list[MaskedArray]:
        """Pixel positions of the observed lines as a list of masked arrays."""
        if self._obs_lines is None:
            return None
        else:
            return [line[:, 0] for line in self._obs_lines]

    @cached_property
    def observed_line_amplitudes(self) -> None | list[MaskedArray]:
        """Amplitudes of the observed lines as a list of masked arrays."""
        if self._obs_lines is None:
            return None
        else:
            return [line[:, 1] for line in self._obs_lines]

    @observed_lines.setter
    def observed_lines(self, line_lists: ArrayLike | list[ArrayLike]):
        if not isinstance(line_lists, Sequence):
            line_lists = [line_lists]
        self._obs_lines = []
        for lst in line_lists:
            self._obs_lines.append(_format_linelist(lst))
        if hasattr(self, "observed_line_locations"):
            del self.observed_line_locations
        if hasattr(self, "observed_line_amplitudes"):
            del self.observed_line_amplitudes

    @property
    def catalog_lines(self) -> None | list[MaskedArray]:
        """Catalog line wavelengths as a list of masked arrays."""
        return self._cat_lines

    @cached_property
    def catalog_line_locations(self) -> None | list[MaskedArray]:
        """Pixel positions of the catalog lines as a list of masked arrays."""
        if self._cat_lines is None:
            return None
        else:
            return [line[:, 0] for line in self._cat_lines]

    @cached_property
    def catalog_line_amplitudes(self) -> None | list[MaskedArray]:
        """Amplitudes of the catalog lines as a list of masked arrays."""
        if self._obs_lines is None:
            return None
        else:
            return [line[:, 1] for line in self._cat_lines]

    @catalog_lines.setter
    def catalog_lines(self, line_lists: ArrayLike | list[ArrayLike]):
        if not isinstance(line_lists, Sequence):
            line_lists = [line_lists]
        self._cat_lines = []
        for lst in line_lists:
            self._cat_lines.append(_format_linelist(lst))
        if hasattr(self, "catalog_line_locations"):
            del self.catalog_line_locations
        if hasattr(self, "catalog_line_amplitudes"):
            del self.catalog_line_amplitudes

    def match_lines(self, max_distance: float = 5) -> None:
        """Match the observed lines to theoretical lines.

        Parameters
        ----------
        max_distance
            The maximum allowed distance between the catalog and observed lines for them to be
            considered a match.
        """

        for iframe, tree in enumerate(self._trees):
            l, ix = tree.query(
                self.solution.p2w(self.observed_line_locations[iframe].data)[:, None],
                distance_upper_bound=max_distance,
            )
            m = np.isfinite(l)

            # Check for observed lines that match a catalog line.
            # Remove all but the nearest match. This isn't an optimal solution,
            # we could also iterate the match by removing the currently matched
            # lines, but this works for now.
            uix, cnt = np.unique(ix[m], return_counts=True)
            if any(n := cnt > 1):
                for i, c in zip(uix[n], cnt[n]):
                    s = ix == i
                    r = np.zeros(c, dtype=bool)
                    r[np.argmin(l[s])] = True
                    m[s] = r

            self._cat_lines[iframe].mask[:, :] = True
            self._cat_lines[iframe].mask[ix[m], :] = False
            self._obs_lines[iframe].mask[:, :] = ~m[:, None]

    def remove_unmatched_lines(self) -> None:
        """Remove unmatched lines from observation and catalog line data."""
        self.observed_lines = [lst.compressed().reshape([-1, 2]) for lst in self._obs_lines]
        self.catalog_lines = [lst.compressed().reshape([-1, 2]) for lst in self._cat_lines]
        self._create_trees()

    def rms(self, space: Literal["pixel", "wavelength"] = "wavelength") -> float:
        """Compute the RMS of the residuals between matched lines in the pixel or wavelength space.

        Parameters
        ----------
        space
            The space in which to calculate the RMS residual. If 'wavelength',
            the calculation is performed in the wavelength space. If 'pixel',
            it is performed in the pixel space. Default is 'wavelength'.

        Returns
        -------
        float
        """
        self.match_lines()
        mpix = np.ma.concatenate(self.observed_line_locations).compressed()
        mwav = np.ma.concatenate(self.catalog_line_locations).compressed()
        if space == "wavelength":
            return np.sqrt(((mwav - self.solution.p2w(mpix)) ** 2).mean())
        elif space == "pixel":
            return np.sqrt(((mpix - self.solution.w2p(mwav)) ** 2).mean())
        else:
            raise ValueError("Space must be either 'pixel' or 'wavelength'")

    def _plot_lines(
        self,
        kind: Literal["observed", "catalog"],
        frames: int | Sequence[int] | None = None,
        axes: Axes | Sequence[Axes] | None = None,
        figsize: tuple[float, float] | None = None,
        plot_labels: bool | Sequence[bool] = True,
        map_x: bool = False,
        label_kwargs: dict | None = None,
    ) -> Figure:
        """
        Plot lines with optional features such as wavelength mapping and label customization.

        Parameters
        ----------
        kind
            Specifies the line list to plot.

        frames
            Frame indices to plot. If None, all frames are plotted.

        axes
            Axes object(s) where the lines should be plotted. If None, new Axes are generated.

        figsize
            Size of the figure to use if creating new Axes. Ignored if axes are provided.

        plot_labels
            Flag(s) indicating whether to display labels for the lines. If a single value is
            provided, it is applied to all frames.

        map_x
            If True, maps the x-axis values between pixel and wavelength space.

        label_kwargs
            Additional keyword arguments to customize the label style.

        Returns
        -------
        Figure
            The Figure object containing the plotted spectral lines.
        """
        largs = dict(backgroundcolor="w", rotation=90, size="small")
        if label_kwargs is not None:
            largs.update(label_kwargs)

        if frames is None:
            frames = np.arange(self.nframes)
        else:
            frames = np.atleast_1d(frames)

        if axes is None:
            fig, axes = subplots(
                frames.size, 1, figsize=figsize, constrained_layout=True, sharex="all"
            )
        elif isinstance(axes, Axes):
            fig = axes.figure
            axes = [axes]
        else:
            fig = axes[0].figure
        axes = np.atleast_1d(axes)

        if isinstance(plot_labels, bool):
            plot_labels = np.full(frames.size, plot_labels, dtype=bool)

        if map_x and self.solution.p2w is None:
            raise ValueError("Cannot map between pixels and wavelengths without a fitted model.")

        if kind == "observed":
            transform = self.solution.pix_to_wav if map_x else lambda x: x
            linelists = self.observed_lines
            spectra = self.arc_spectra
            lc = "C0"
        else:
            transform = self.solution.wav_to_pix if map_x else lambda x: x
            linelists = self.catalog_lines
            spectra = None
            lc = "C1"

        ypad = 1.3
        labels = []
        for iframe, (ax, frame) in enumerate(zip(axes, frames)):
            if spectra is not None:
                spc = self.arc_spectra[iframe]
                vmax = np.nanmax(spc.flux.value)
                ax.plot(transform(spc.spectral_axis.value), spc.flux.value / vmax, "k")
            else:
                vmax = 1.0

            if linelists is not None:
                labels.append([])

                # Loop over individual lines in the line list.
                for i in range(linelists[iframe].shape[0]):
                    c, a = linelists[iframe].data[i]
                    ls = "-" if linelists[iframe].mask[i, 0] == 0 else ":"

                    ax.plot(transform([c, c]), [a / vmax + 0.1, 1.27], c=lc, ls=ls, zorder=-100)
                    if plot_labels[iframe]:
                        lloc = transform(c)
                        labels[-1].append(
                            ax.text(
                                lloc,
                                ypad,
                                np.round(lloc, 4 - 1 - int(np.floor(np.log10(lloc)))),
                                ha="center",
                                va="top",
                                **largs,
                            )
                        )
                        labels[-1][-1].set_clip_on(True)
                        labels[-1][-1].zorder = a

        if (kind == "observed" and not map_x) or (kind == "catalog" and map_x):
            xlabel = "Pixel"
        else:
            xlabel = f"Wavelength {self._unit_str}"

        if kind == "catalog":
            axes[0].xaxis.set_label_position("top")
            axes[0].xaxis.tick_top()
            setp(axes[0], xlabel=xlabel)
            for ax in axes[1:]:
                ax.set_xticklabels([])
        else:
            setp(axes[-1], xlabel=xlabel)
            for ax in axes[:-1]:
                ax.set_xticklabels([])

        xlims = np.array([ax.get_xlim() for ax in axes])
        setp(axes, xlim=(xlims[:, 0].min(), xlims[:, 1].max()), yticks=[])

        if linelists is not None:
            fig.canvas.draw()
            for i in range(len(frames)):
                if plot_labels[i]:

                    # Calculate the label bounding box upper limits and adjust the y-axis limits.
                    tr_to_data = axes[i].transData.inverted()
                    ymax = -np.inf
                    for lb in labels[i]:
                        ymax = max(ymax, tr_to_data.transform(lb.get_window_extent().p1)[1])
                    setp(axes[i], ylim=(-0.04, ymax * 1.06))

                    # Remove the overlapping labels prioritizing the high-amplitude lines.
                    _unclutter_text_boxes(labels[i])

        return fig

    def plot_catalog_lines(
        self,
        frames: int | Sequence[int] | None = None,
        axes: Axes | Sequence[Axes] | None = None,
        figsize: tuple[float, float] | None = None,
        plot_labels: bool | Sequence[bool] = True,
        map_to_pix: bool = False,
        label_kwargs: dict | None = None,
    ) -> Figure:
        """Plot the catalog lines.

        Parameters
        ----------
        frames
            Specifies the frames to be plotted. If an integer, only one frame is plotted.
            If a sequence, the specified frames are plotted. If None, default selection
            or all frames are plotted.

        axes
            The matplotlib axes where catalog data will be plotted. If provided, the function
            will plot on these axes. If None, new axes will be created.

        figsize
            Specifies the dimensions of the figure as (width, height). If None, the default
            dimensions are used.

        plot_labels
            If True, the numerical values associated with the catalog data will be displayed
            in the plot. If False, only the graphical representation of the lines will be shown.

        map_to_pix
            Indicates whether the catalog data should be mapped to pixel coordinates
            before plotting. If True, the data is converted to pixel coordinates.

        label_kwargs
            Specifies the keyword arguments for the line label text objects.

        Returns
        -------
        Figure
            The matplotlib figure containing the plotted catalog lines.

        """
        return self._plot_lines(
            "catalog",
            frames=frames,
            axes=axes,
            figsize=figsize,
            plot_labels=plot_labels,
            map_x=map_to_pix,
            label_kwargs=label_kwargs,
        )

    def plot_observed_lines(
        self,
        frames: int | Sequence[int] | None = None,
        axes: Axes | Sequence[Axes] | None = None,
        figsize: tuple[float, float] | None = None,
        plot_labels: bool | Sequence[bool] = True,
        map_to_wav: bool = False,
        label_kwargs: dict | None = None,
    ) -> Figure:
        """Plot observed spectral lines for the given arc spectra.

        Parameters
        ----------
        frames
            Specifies the frame(s) for which the plot is to be generated. If None, all frames
            are plotted. When an integer is provided, a single frame is used. For a sequence
            of integers, multiple frames are plotted.

        axes
            Axes object(s) to plot the spectral lines on. If None, new axes are created.

        figsize
            Dimensions of the figure to be created, specified as a tuple (width, height). Ignored
            if ``axes`` is provided.

        plot_labels
            If True, plots the numerical values of the observed lines at their respective
            locations on the graph.

        map_to_wav
            Determines whether to map the x-axis values to wavelengths.

        label_kwargs
            Specifies the keyword arguments for the line label text objects.

        Returns
        -------
        Figure
            The matplotlib figure containing the observed lines plot.
        """

        fig = self._plot_lines(
            "observed",
            frames=frames,
            axes=axes,
            figsize=figsize,
            plot_labels=plot_labels,
            map_x=map_to_wav,
            label_kwargs=label_kwargs,
        )

        for ax in fig.axes:
            ax.autoscale(True, "x", tight=True)
        return fig

    def plot_fit(
        self,
        frames: Sequence[int] | int | None = None,
        figsize: tuple[float, float] | None = None,
        plot_labels: bool = True,
        obs_to_wav: bool = False,
        cat_to_pix: bool = False,
        label_kwargs: dict | None = None,
    ) -> Figure:
        """Plot the fitted catalog and observed lines for the specified arc spectra.

        Parameters
        ----------
        frames
            The indices of the frames to plot. If `None`, all frames from 0 to
            ``self.nframes - 1`` are plotted.

        figsize
            Defines the width and height of the figure in inches. If `None`, the
            default size is used.

        plot_labels
            If `True`, print line locations over the plotted lines. Can also be a list with
            the same length as ``frames``.

        obs_to_wav
            If `True`, transform the x-axis of observed lines to the wavelength domain
            using `self._p2w`, if available.

        cat_to_pix
            If `True`, transforms catalog data points to pixel values before plotting.

        label_kwargs
            Specifies the keyword arguments for the line label text objects.

        Returns
        -------
        matplotlib.figure.Figure
            The figure object containing the generated subplots.
        """
        if frames is None:
            frames = np.arange(self.nframes)
        else:
            frames = np.atleast_1d(frames)

        fig, axs = subplots(2 * frames.size, 1, constrained_layout=True, figsize=figsize)
        self.plot_catalog_lines(
            frames,
            axs[0::2],
            plot_labels=plot_labels,
            map_to_pix=cat_to_pix,
            label_kwargs=label_kwargs,
        )
        self.plot_observed_lines(
            frames,
            axs[1::2],
            plot_labels=plot_labels,
            map_to_wav=obs_to_wav,
            label_kwargs=label_kwargs,
        )

        xlims = np.array([ax.get_xlim() for ax in axs[::2]])
        if obs_to_wav:
            setp(axs, xlim=(xlims[:, 0].min(), xlims[:, 1].max()))
        else:
            setp(axs[::2], xlim=(xlims[:, 0].min(), xlims[:, 1].max()))

        setp(axs[0], yticks=[], xlabel=f"Wavelength [{self._unit_str}]")
        for ax in axs[1:-1]:
            ax.set_xlabel("")
            ax.set_xticklabels("")

        axs[0].xaxis.set_label_position("top")
        axs[0].xaxis.tick_top()
        return fig

    def plot_residuals(
        self,
        ax: Axes | None = None,
        space: Literal["pixel", "wavelength"] = "wavelength",
        figsize: tuple[float, float] | None = None,
    ) -> Figure:
        """Plot the residuals of pixel-to-wavelength or wavelength-to-pixel transformation.

        Parameters
        ----------
        ax
            Matplotlib Axes object to plot on. If None, a new figure and axes are created.

        space
            The reference space used for plotting residuals. Options are 'pixel' for residuals
            in pixel space or 'wavelength' for residuals in wavelength space.

        figsize
            The size of the figure in inches, if a new figure is created.

        Returns
        -------
        matplotlib.figure.Figure
        """
        if ax is None:
            fig, ax = subplots(figsize=figsize, constrained_layout=True)
        else:
            fig = ax.figure

        self.match_lines()
        mpix = np.ma.concatenate(self.observed_line_locations).compressed()
        mwav = np.ma.concatenate(self.catalog_line_locations).compressed()

        if space == "wavelength":
            twav = self.solution.pix_to_wav(mpix)
            ax.plot(mwav, mwav - twav, ".")
            ax.text(
                0.98,
                0.95,
                f"RMS = {np.sqrt(((mwav - twav) ** 2).mean()):4.2f} {self._unit_str}",
                transform=ax.transAxes,
                ha="right",
                va="top",
            )
            setp(
                ax,
                xlabel=f"Wavelength [{self._unit_str}]",
                ylabel=f"Residuals [{self._unit_str}]",
            )
        elif space == "pixel":
            tpix = self.solution.wav_to_pix(mwav)
            ax.plot(mpix, mpix - tpix, ".")
            ax.text(
                0.98,
                0.95,
                f"RMS = {np.sqrt(((mpix - tpix) ** 2).mean()):4.2f} pix",
                transform=ax.transAxes,
                ha="right",
                va="top",
            )
            setp(ax, xlabel="Pixel", ylabel="Residuals [pix]")
        else:
            raise ValueError("Invalid space specified for plotting residuals.")
        ax.axhline(0, c="k", lw=1, ls="--")
        return fig



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import warnings
from dataclasses import dataclass, field

import numpy as np
from astropy import units as u
from astropy.utils.decorators import deprecated_attribute

from specreduce.compat import SPECUTILS_LT_2, Spectrum
from specreduce.core import _ImageParser, MaskingOption, ImageLike
from specreduce.extract import _ap_weight_image
from specreduce.tracing import Trace, FlatTrace

__all__ = ["Background"]


@dataclass
class Background(_ImageParser):
    """
    Determine the background from an image for subtraction.

    Example: ::

        trace = FlatTrace(image, trace_pos)
        bg = Background.two_sided(image, trace, bkg_sep, width=bkg_width)
        subtracted_image = image - bg

    Parameters
    ----------
    image
        Image with 2-D spectral image data
    traces : List, `specreduce.tracing.Trace`, int, float
        Individual or list of trace object(s) (or integers/floats to define
        FlatTraces) to extract the background. If None, a ``FlatTrace`` at the
        center of the image (according to ``disp_axis``) will be used.
    width
        Width of extraction aperture in pixels.
    statistic
        Statistic to use when computing the background.  ``average`` will
        account for partial pixel weights, ``median`` will include all partial
        pixels.
    disp_axis
        Dispersion axis.
    crossdisp_axis
        Cross-dispersion axis.
    mask_treatment
        Specifies how to handle masked or non-finite values in the input image.
        The accepted values are:

        - ``apply``: The image remains unchanged, and any existing mask is combined\
            with a mask derived from non-finite values.
        - ``ignore``: The image remains unchanged, and any existing mask is dropped.
        - ``propagate``: The image remains unchanged, and any masked or non-finite pixel\
            causes the mask to extend across the entire cross-dispersion axis.
        - ``zero_fill``: Pixels that are either masked or non-finite are replaced with 0.0,\
            and the mask is dropped.
        - ``nan_fill``:  Pixels that are either masked or non-finite are replaced with nan,\
            and the mask is dropped.
        - ``apply_mask_only``: The  image and mask are left unmodified.
        - ``apply_nan_only``: The  image is left unmodified, the old mask is dropped, and a\
            new mask is created based on non-finite values.

"""

    # required so numpy won't call __rsub__ on individual elements
    # https://stackoverflow.com/a/58409215
    __array_ufunc__ = None

    image: ImageLike
    traces: list = field(default_factory=list)
    width: float = 5
    statistic: str = "average"
    disp_axis: int = 1
    crossdisp_axis: int = 0
    mask_treatment: MaskingOption = "apply"
    _valid_mask_treatment_methods = (
        "apply",
        "ignore",
        "propagate",
        "zero_fill",
        "nan_fill",
        "apply_mask_only",
        "apply_nan_only",
    )

    # TO-DO: update bkg_array with Spectrum alternative (is bkg_image enough?)
    bkg_array = deprecated_attribute("bkg_array", "1.3")

    def __post_init__(self):
        self.image = self._parse_image(
            self.image, disp_axis=self.disp_axis, mask_treatment=self.mask_treatment
        )

        # always work with masked array, even if there is no masked
        # or non-finite data, in case padding is needed. if not, mask will be
        # dropped at the end and a regular array will be returned.
        img = np.ma.masked_array(self.image.data, self.image.mask)

        if self.width < 0:
            raise ValueError("width must be positive")
        if self.width == 0:
            self._bkg_array = np.zeros(self.image.shape[self.disp_axis])
            return

        self._set_traces()

        bkg_wimage = np.zeros_like(self.image.data, dtype=np.float64)
        for trace in self.traces:
            # note: ArrayTrace can have masked values, but if it does a MaskedArray
            # will be returned so this should be reflected in the window size here
            # (i.e, np.nanmax is not required.)
            windows_max = trace.trace.data.max() + self.width / 2
            windows_min = trace.trace.data.min() - self.width / 2

            if windows_max > self.image.shape[self.crossdisp_axis]:
                warnings.warn(
                    "background window extends beyond image boundaries "
                    + f"({windows_max} >= {self.image.shape[self.crossdisp_axis]})"
                )
            if windows_min < 0:
                warnings.warn(
                    "background window extends beyond image boundaries " + f"({windows_min} < 0)"
                )

            # pass trace.trace.data to ignore any mask on the trace
            bkg_wimage += _ap_weight_image(
                trace, self.width, self.disp_axis, self.crossdisp_axis, self.image.shape
            )

        if np.any(bkg_wimage > 1):
            raise ValueError("background regions overlapped")
        if np.any(np.sum(bkg_wimage, axis=self.crossdisp_axis) == 0):
            raise ValueError(
                "background window does not remain in bounds across entire dispersion axis"
            )  # noqa
        # check if image contained within background window is fully-nonfinite and raise an error
        if np.all(img.mask[bkg_wimage > 0]):
            raise ValueError(
                "Image is fully masked within background window determined by `width`."
            )  # noqa

        if self.statistic == "median":
            # make it clear in the expose image that partial pixels are fully-weighted
            bkg_wimage[bkg_wimage > 0] = 1

        self.bkg_wimage = bkg_wimage

        if self.statistic == "average":
            self._bkg_array = np.ma.average(img, weights=self.bkg_wimage, axis=self.crossdisp_axis)

        elif self.statistic == "median":
            # combine where background weight image is 0 with image masked (which already
            # accounts for non-finite data that wasn't already masked)
            img.mask = np.logical_or(self.bkg_wimage == 0, self.image.mask)
            self._bkg_array = np.ma.median(img, axis=self.crossdisp_axis)
        else:
            raise ValueError("statistic must be 'average' or 'median'")

    def _set_traces(self):
        """Determine `traces` from input. If an integer/float or list if int/float
        is passed in, use these to construct FlatTrace objects. These values
        must be positive. If None (which is initialized to an empty list),
        construct a FlatTrace using the center of image (according to disp.
        axis). Otherwise, any Trace object or list of Trace objects can be
        passed in."""

        if self.traces == []:
            # assume a flat trace at the image center if nothing is passed in.
            trace_pos = self.image.shape[self.disp_axis] / 2.0
            self.traces = [FlatTrace(self.image, trace_pos)]

        if isinstance(self.traces, Trace):
            # if just one trace, turn it into iterable.
            self.traces = [self.traces]
            return

        # finally, if float/int is passed in convert to FlatTrace(s)
        if isinstance(self.traces, (float, int)):  # for a single number
            self.traces = [self.traces]

        if np.all([isinstance(x, (float, int)) for x in self.traces]):
            self.traces = [FlatTrace(self.image, trace_pos) for trace_pos in self.traces]
            return

        else:
            if not np.all([isinstance(x, Trace) for x in self.traces]):
                raise ValueError(
                    "`traces` must be a `Trace` object or list of "
                    "`Trace` objects, a number or list of numbers to "
                    "define FlatTraces, or None to use a FlatTrace in "
                    "the middle of the image."
                )

    @classmethod
    def two_sided(cls, image, trace_object, separation, **kwargs):
        """
        Determine the background from an image for subtraction centered around
        an input trace.


        Example: ::

            trace = FitTrace(image, guess=trace_pos)
            bg = Background.two_sided(image, trace, bkg_sep, width=bkg_width)

        Parameters
        ----------
        image : `~astropy.nddata.NDData`-like or array-like
            Image with 2-D spectral image data. Assumes cross-dispersion
            (spatial) direction is axis 0 and dispersion (wavelength)
            direction is axis 1.
        trace_object: `~specreduce.tracing.Trace`
            estimated trace of the spectrum to center the background traces
        separation: float
            separation from ``trace_object`` for the background regions
        width : float
            width of each background aperture in pixels
        statistic: string
            statistic to use when computing the background.  'average' will
            account for partial pixel weights, 'median' will include all partial
            pixels.
        disp_axis : int
            dispersion axis
        crossdisp_axis : int
            cross-dispersion axis
        mask_treatment : string
            The method for handling masked or non-finite data. Choice of ``filter``,
            ``omit`, or ``zero_fill``. If `filter` is chosen, masked/non-finite data
            will be filtered during the fit to each bin/column (along disp. axis) to
            find the peak. If ``omit`` is chosen, columns along disp_axis with any
            masked/non-finite data values will be fully masked (i.e, 2D mask is
            collapsed to 1D and applied). If ``zero_fill`` is chosen, masked/non-finite
            data will be replaced with 0.0 in the input image, and the mask will then
            be dropped. For all three options, the input mask (optional on input
            NDData object) will be combined with a mask generated from any non-finite
            values in the image data.
        """

        image = _ImageParser._get_data_from_image(image) if image is not None else cls.image
        kwargs["traces"] = [trace_object - separation, trace_object + separation]
        return cls(image=image, **kwargs)

    @classmethod
    def one_sided(cls, image, trace_object, separation, **kwargs):
        """
        Determine the background from an image for subtraction above
        or below an input trace.

        Example: ::

            trace = FitTrace(image, guess=trace_pos)
            bg = Background.one_sided(image, trace, bkg_sep, width=bkg_width)

        Parameters
        ----------
        image : `~astropy.nddata.NDData`-like or array-like
            Image with 2-D spectral image data. Assumes cross-dispersion
            (spatial) direction is axis 0 and dispersion (wavelength)
            direction is axis 1.
        trace_object: `~specreduce.tracing.Trace`
            estimated trace of the spectrum to center the background traces
        separation: float
            separation from ``trace_object`` for the background, positive will be
            above the trace, negative below.
        width : float
            width of each background aperture in pixels
        statistic: string
            statistic to use when computing the background.  'average' will
            account for partial pixel weights, 'median' will include all partial
            pixels.
        disp_axis : int
            dispersion axis
        crossdisp_axis : int
            cross-dispersion axis
        mask_treatment : string
            The method for handling masked or non-finite data. Choice of ``filter``,
            ``omit``, or ``zero_fill``. If `filter` is chosen, masked/non-finite data
            will be filtered during the fit to each bin/column (along disp. axis) to
            find the peak. If ``omit`` is chosen, columns along disp_axis with any
            masked/non-finite data values will be fully masked (i.e, 2D mask is
            collapsed to 1D and applied). If ``zero_fill`` is chosen, masked/non-finite
            data will be replaced with 0.0 in the input image, and the mask will then
            be dropped. For all three options, the input mask (optional on input
            NDData object) will be combined with a mask generated from any non-finite
            values in the image data.
        """
        image = _ImageParser._get_data_from_image(image) if image is not None else cls.image
        kwargs["traces"] = [trace_object + separation]
        return cls(image=image, **kwargs)

    def bkg_image(self, image=None):
        """
        Expose the background tiled to the dimension of ``image``.

        Parameters
        ----------
        image : `~astropy.nddata.NDData`-like or array-like, optional
            Image with 2-D spectral image data. Assumes cross-dispersion
            (spatial) direction is axis 0 and dispersion (wavelength)
            direction is axis 1. If None, will extract the background
            from ``image`` used to initialize the class. [default: None]

        Returns
        -------
        spec : `~specutils.Spectrum1D`
            Spectrum object with same shape as ``image``.
        """
        image = self._parse_image(image)
        arr = np.tile(self._bkg_array, (image.shape[0], 1))
        if SPECUTILS_LT_2:
            kwargs = {}
        else:
            kwargs = {"spectral_axis_index": arr.ndim - 1}
        return Spectrum(
            arr * image.unit,
            spectral_axis=image.spectral_axis, **kwargs
        )

    def bkg_spectrum(self, image=None, bkg_statistic=None):
        """
        Expose the 1D spectrum of the background.

        Parameters
        ----------
        image : `~astropy.nddata.NDData`-like or array-like, optional
            Image with 2-D spectral image data. Assumes cross-dispersion
            (spatial) direction is axis 0 and dispersion (wavelength)
            direction is axis 1. If None, will extract the background
            from ``image`` used to initialize the class. [default: None]

        Returns
        -------
        spec : `~specutils.Spectrum1D`
            The background 1-D spectrum, with flux expressed in the same
            units as the input image (or DN if none were provided).
        """
        if bkg_statistic is not None:
            warnings.warn("'bkg_statistic' is deprecated and will be removed in a future release. "
                          "Please use the 'statistic' argument in the Background initializer instead.",  # noqa
                          DeprecationWarning,)

        return Spectrum(self._bkg_array * self.image.unit, self.image.spectral_axis)

    def sub_image(self, image=None):
        """
        Subtract the computed background from ``image``.

        Parameters
        ----------
        image : nddata-compatible image or None
            image with 2-D spectral image data.  If None, will extract
            the background from ``image`` used to initialize the class.

        Returns
        -------
        spec : `~specutils.Spectrum1D`
            Spectrum object with same shape as ``image``.
        """
        image = self._parse_image(image)

        if not SPECUTILS_LT_2:
            return image - self.bkg_image(image)

        # a compare_wcs argument is needed for Spectrum.subtract() in order to
        # avoid a TypeError from SpectralCoord when image's spectral axis is in
        # pixels. it is not needed when image's spectral axis has physical units
        kwargs = {"compare_wcs": None} if image.spectral_axis.unit == u.pix else {}

        # https://docs.astropy.org/en/stable/nddata/mixins/ndarithmetic.html
        return image.subtract(self.bkg_image(image), **kwargs)

    def sub_spectrum(self, image=None):
        """
        Expose the 1D spectrum of the background-subtracted image.

        Parameters
        ----------
        image : nddata-compatible image or None
            image with 2-D spectral image data.  If None, will extract
            the background from ``image`` used to initialize the class.

        Returns
        -------
        spec : `~specutils.Spectrum1D`
            The background 1-D spectrum, with flux expressed in the same
            units as the input image (or u.DN if none were provided) and
            the spectral axis expressed in pixel units.
        """
        sub_image = self.sub_image(image=image)

        try:
            return sub_image.collapse(np.nansum, axis=self.crossdisp_axis)
        except u.UnitTypeError:
            # can't collapse with a spectral axis in pixels because
            # SpectralCoord only allows frequency/wavelength equivalent units...
            ext1d = np.nansum(sub_image.flux, axis=self.crossdisp_axis)
            return Spectrum(ext1d, spectral_axis=sub_image.spectral_axis)

    def __rsub__(self, image):
        """
        Subtract the background from an image.
        """
        return self.sub_image(image)



# Licensed under a 3-clause BSD style license - see ../../licenses/LICENSE.rst
import warnings
from typing import Sequence

import numpy as np
from astropy import units as u
from astropy.modeling import models, Model
from astropy.nddata import CCDData
from astropy.stats import gaussian_fwhm_to_sigma
from astropy.wcs import WCS

from specreduce.calibration_data import load_pypeit_calibration_lines

__all__ = [
    'make_2d_trace_image',
    'make_2d_arc_image',
    'make_2d_spec_image'
]


def make_2d_trace_image(
    nx: int = 3000,
    ny: int = 1000,
    background: int | float = 5,
    trace_center: int | float | None = None,
    trace_order: int = 3,
    trace_coeffs: dict[str, int | float] = {'c0': 0, 'c1': 50, 'c2': 100},
    profile: Model = models.Moffat1D(amplitude=10, alpha=0.1),
    add_noise: bool = True
) -> CCDData:
    """
    Create synthetic 2D spectroscopic image with a single source. The spatial (y-axis) position
    of the source along the dispersion (x-axis) direction is modeled using a Chebyshev polynomial.
    The flux units are counts and the noise is modeled as Poisson.

    Parameters
    ----------
    nx : Size of image in X axis which is assumed to be the dispersion axis

    ny : Size of image in Y axis which is assumed to be the spatial axis

    background : Level of constant background in counts

    trace_center : Zeropoint of the trace. If None, then use center of Y (spatial) axis.

    trace_order : Order of the Chebyshev polynomial used to model the source's trace

    trace_coeffs : Dict containing the Chebyshev polynomial coefficients to use in the trace model

    profile : Model to use for the source's spatial profile

    add_noise : If True, add Poisson noise to the image


    Returns
    -------
    ccd_im : CCDData instance containing synthetic 2D spectroscopic image
    """
    x = np.arange(nx)
    y = np.arange(ny)
    xx, yy = np.meshgrid(x, y)

    if trace_center is None:
        trace_center = ny / 2

    trace_mod = models.Chebyshev1D(degree=trace_order, **trace_coeffs)
    trace = yy - trace_center + trace_mod(xx/nx)
    z = background + profile(trace)

    if add_noise:
        from photutils.datasets import apply_poisson_noise
        trace_image = apply_poisson_noise(z)
    else:
        trace_image = z

    ccd_im = CCDData(trace_image, unit=u.count)

    return ccd_im


def make_2d_arc_image(
    nx: int = 3000,
    ny: int = 1000,
    wcs: WCS | None = None,
    extent: Sequence[int | float] = (3500, 7000),
    wave_unit: u.Unit = u.Angstrom,
    wave_air: bool = False,
    background: int | float = 5,
    line_fwhm: float = 5.,
    linelists: list[str] = ['HeI'],
    amplitude_scale: float = 1.,
    tilt_func: Model = models.Legendre1D(degree=0),
    add_noise: bool = True
) -> CCDData:
    """
    Create synthetic 2D spectroscopic image of reference emission lines, e.g. a calibration
    arc lamp. Currently, linelists from ``pypeit`` are supported and are selected by string or
    list of strings that is passed to `~specreduce.calibration_data.load_pypeit_calibration_lines`.
    If a ``wcs`` is not provided, one is created using ``extent`` and ``wave_unit`` with
    dispersion along the X axis.

    Parameters
    ----------
    nx : Size of image in X axis which is assumed to be the dispersion axis

    ny : Size of image in Y axis which is assumed to be the spatial axis

    wcs : 2D WCS to apply to the image. Must have a spectral axis defined along with
        appropriate spectral wavelength units.

    extent : If ``wcs`` is not provided, this defines the beginning and end wavelengths
        of the dispersion axis.

    wave_unit : If ``wcs`` is not provided, this defines the wavelength units of the
        dispersion axis.

    wave_air : If True, convert the vacuum wavelengths used by ``pypeit`` to air wavelengths.

    background : Level of constant background in counts

    line_fwhm : Gaussian FWHM of the emission lines in pixels

    linelists : Specification for linelists to load from ``pypeit``

    amplitude_scale : Scale factor to apply to amplitudes provided in the linelists

    tilt_func : The tilt function to apply along the cross-dispersion axis to simulate
        tilted or curved emission lines.
    add_noise : If True, add Poisson noise to the image; requires ``photutils`` to be installed.

    Returns
    -------
    ccd_im : CCDData instance containing synthetic 2D spectroscopic image

    Examples
    --------
    This is an example of modeling a spectrograph whose output is curved in the
    cross-dispersion direction:

    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        import numpy as np
        from astropy.modeling import models
        import astropy.units as u
        from specreduce.utils.synth_data import make_2d_arc_image

        model_deg2 = models.Legendre1D(degree=2, c0=50, c1=0, c2=100)
        im = make_2d_arc_image(
            linelists=['HeI', 'ArI', 'ArII'],
            line_fwhm=3,
            tilt_func=model_deg2
        )
        fig = plt.figure(figsize=(10, 6))
        plt.imshow(im)

    The FITS WCS standard implements ideal world coordinate functions based on the physics
    of simple dispersers. This is described in detail by Paper III,
    https://www.aanda.org/articles/aa/pdf/2006/05/aa3818-05.pdf. This can be used to model a
    non-linear dispersion relation based on the properties of a spectrograph. This example
    recreates Figure 5 in that paper using a spectrograph with a 450 lines/mm volume phase
    holographic grism. Standard gratings only use the first three ``PV`` terms:

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        from astropy.wcs import WCS
        import astropy.units as u
        from specreduce.utils.synth_data import make_2d_arc_image

        non_linear_header = {
            'CTYPE1': 'AWAV-GRA',  # Grating dispersion function with air wavelengths
            'CUNIT1': 'Angstrom',  # Dispersion units
            'CRPIX1': 719.8,       # Reference pixel [pix]
            'CRVAL1': 7245.2,      # Reference value [Angstrom]
            'CDELT1': 2.956,       # Linear dispersion [Angstrom/pix]
            'PV1_0': 4.5e5,        # Grating density [1/m]
            'PV1_1': 1,            # Diffraction order
            'PV1_2': 27.0,         # Incident angle [deg]
            'PV1_3': 1.765,        # Reference refraction
            'PV1_4': -1.077e6,     # Refraction derivative [1/m]
            'CTYPE2': 'PIXEL',     # Spatial detector coordinates
            'CUNIT2': 'pix',       # Spatial units
            'CRPIX2': 1,           # Reference pixel
            'CRVAL2': 0,           # Reference value
            'CDELT2': 1            # Spatial units per pixel
        }

        linear_header = {
            'CTYPE1': 'AWAV',  # Grating dispersion function with air wavelengths
            'CUNIT1': 'Angstrom',  # Dispersion units
            'CRPIX1': 719.8,       # Reference pixel [pix]
            'CRVAL1': 7245.2,      # Reference value [Angstrom]
            'CDELT1': 2.956,       # Linear dispersion [Angstrom/pix]
            'CTYPE2': 'PIXEL',     # Spatial detector coordinates
            'CUNIT2': 'pix',       # Spatial units
            'CRPIX2': 1,           # Reference pixel
            'CRVAL2': 0,           # Reference value
            'CDELT2': 1            # Spatial units per pixel
        }

        non_linear_wcs = WCS(non_linear_header)
        linear_wcs = WCS(linear_header)

        # this re-creates Paper III, Figure 5
        pix_array = 200 + np.arange(1400)
        nlin = non_linear_wcs.spectral.pixel_to_world(pix_array)
        lin = linear_wcs.spectral.pixel_to_world(pix_array)
        resid = (nlin - lin).to(u.Angstrom)
        plt.plot(pix_array, resid)
        plt.xlabel("Pixel")
        plt.ylabel("Correction (Angstrom)")
        plt.show()

        nlin_im = make_2d_arc_image(
            nx=600,
            ny=512,
            linelists=['HeI', 'NeI'],
            line_fwhm=3,
            wave_air=True,
            wcs=non_linear_wcs
        )
        lin_im = make_2d_arc_image(
            nx=600,
            ny=512,
            linelists=['HeI', 'NeI'],
            line_fwhm=3,
            wave_air=True,
            wcs=linear_wcs
        )

        # subtracting the linear simulation from the non-linear one shows how the
        # positions of lines diverge between the two cases
        plt.imshow(nlin_im.data - lin_im.data)
        plt.show()

    """
    if wcs is None:
        if extent is None:
            raise ValueError("Must specify either a wavelength extent or a WCS.")
        if len(extent) != 2:
            raise ValueError("Wavelength extent must be of length 2.")
        if u.get_physical_type(wave_unit) != 'length':
            raise ValueError("Wavelength unit must be a length unit.")
        wcs = WCS(naxis=2)
        wcs.wcs.ctype[0] = 'WAVE'
        wcs.wcs.ctype[1] = 'PIXEL'
        wcs.wcs.cunit[0] = wave_unit
        wcs.wcs.cunit[1] = u.pixel
        wcs.wcs.crval[0] = extent[0]
        wcs.wcs.cdelt[0] = (extent[1] - extent[0]) / nx
        wcs.wcs.crval[1] = 0
        wcs.wcs.cdelt[1] = 1
    else:
        if wcs.spectral.naxis != 1:
            raise ValueError("Provided WCS must have a spectral axis.")
        if wcs.naxis != 2:
            raise ValueError("WCS must have NAXIS=2 for a 2D image.")

    x = np.arange(nx)
    y = np.arange(ny)
    xx, yy = np.meshgrid(x, y)

    is_spectral = [a['coordinate_type'] == "spectral" for a in wcs.get_axis_types()]
    if is_spectral[0]:
        disp_axis = 0
    else:
        disp_axis = 1

    if tilt_func is not None:
        if not isinstance(
            tilt_func,
            (models.Legendre1D, models.Chebyshev1D, models.Polynomial1D, models.Hermite1D)
        ):
            raise ValueError(
                "The only tilt functions currently supported are 1D polynomials "
                "from astropy.models."
            )

        if disp_axis == 0:
            xx = xx + tilt_func((yy - ny/2)/ny)
        else:
            yy = yy + tilt_func((xx - nx/2)/nx)

    z = background + np.zeros((ny, nx))

    linelist = load_pypeit_calibration_lines(linelists, wave_air=wave_air)

    if linelist is not None:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="No observer defined on WCS.*")
            line_disp_positions = wcs.spectral.world_to_pixel(linelist['wavelength'])

        line_sigma = gaussian_fwhm_to_sigma * line_fwhm
        for line_pos, ampl in zip(line_disp_positions, linelist['amplitude']):
            line_mod = models.Gaussian1D(
                amplitude=ampl * amplitude_scale,
                mean=line_pos,
                stddev=line_sigma
            )
            if disp_axis == 0:
                z += line_mod(xx)
            else:
                z += line_mod(yy)

    if add_noise:
        from photutils.datasets import apply_poisson_noise
        arc_image = apply_poisson_noise(z)
    else:
        arc_image = z

    ccd_im = CCDData(arc_image, unit=u.count, wcs=wcs)

    return ccd_im


def make_2d_spec_image(
    nx: int = 3000,
    ny: int = 1000,
    wcs: WCS | None = None,
    extent: Sequence[int | float] = (6500, 9500),
    wave_unit: u.Unit = u.Angstrom,
    wave_air: bool = False,
    background: int | float = 5,
    line_fwhm: float = 5.,
    linelists: list[str] = ['OH_GMOS'],
    amplitude_scale: float = 1.,
    tilt_func: Model = models.Legendre1D(degree=0),
    trace_center: int | float | None = None,
    trace_order: int = 3,
    trace_coeffs: dict[str, int | float] = {'c0': 0, 'c1': 50, 'c2': 100},
    source_profile: Model = models.Moffat1D(amplitude=10, alpha=0.1),
    add_noise: bool = True
) -> CCDData:
    """
    Make a synthetic 2D spectrum image containing both emission lines and
    a trace of a continuum source.

    Parameters
    ----------
    nx : Number of pixels in the dispersion direction.

    ny : Number of pixels in the spatial direction.

    wcs : 2D WCS to apply to the image. Must have a spectral axis defined along with
        appropriate spectral wavelength units.

    extent : If ``wcs`` is not provided, this defines the beginning and end wavelengths
        of the dispersion axis.

    wave_unit : If ``wcs`` is not provided, this defines the wavelength units of the
        dispersion axis.

    wave_air : If True, convert the vacuum wavelengths used by ``pypeit`` to air wavelengths.

    background : Constant background level in counts.

    line_fwhm : Gaussian FWHM of the emission lines in pixels

    linelists : Specification for linelists to load from ``pypeit``

    amplitude_scale : Scale factor to apply to amplitudes provided in the linelists

    tilt_func : The tilt function to apply along the cross-dispersion axis to simulate
        tilted or curved emission lines.

    trace_center : Zeropoint of the trace. If None, then use center of Y (spatial) axis.

    trace_order : Order of the Chebyshev polynomial used to model the source's trace

    trace_coeffs : Dict containing the Chebyshev polynomial coefficients to use in the trace model

    source_profile : Model to use for the source's spatial profile

    add_noise : If True, add Poisson noise to the image; requires ``photutils`` to be installed.


    Returns
    -------
    ccd_im : CCDData instance containing synthetic 2D spectroscopic image
    """
    arc_image = make_2d_arc_image(
        nx=nx,
        ny=ny,
        wcs=wcs,
        extent=extent,
        wave_unit=wave_unit,
        wave_air=wave_air,
        background=0,
        line_fwhm=line_fwhm,
        linelists=linelists,
        amplitude_scale=amplitude_scale,
        tilt_func=tilt_func,
        add_noise=False
    )

    trace_image = make_2d_trace_image(
        nx=nx,
        ny=ny,
        background=0,
        trace_center=trace_center,
        trace_order=trace_order,
        trace_coeffs=trace_coeffs,
        profile=source_profile,
        add_noise=False
    )

    spec_image = arc_image.data + trace_image.data + background

    if add_noise:
        from photutils.datasets import apply_poisson_noise
        spec_image = apply_poisson_noise(spec_image)

    ccd_im = CCDData(spec_image, unit=u.count, wcs=arc_image.wcs)

    return ccd_im



import warnings
from typing import Sequence

import astropy.units as u
import numpy as np
from astropy.stats import gaussian_fwhm_to_sigma, gaussian_sigma_to_fwhm
from astropy.modeling import models
from astropy.table import QTable
from astropy.wcs import WCS as astropy_WCS
from gwcs.wcs import WCS as gWCS
from specutils.fitting import find_lines_threshold, fit_lines

from specreduce.compat import Spectrum

__all__ = [
    "find_arc_lines",
    "match_lines_wcs"
]


def find_arc_lines(
    spectrum: Spectrum,
    fwhm: float | u.Quantity = 5.0 * u.pix,
    window: float = 3.0,
    noise_factor: float = 5.0
) -> QTable:
    """
    Find arc lines in a spectrum using `~specutils.fitting.find_lines_threshold` and
    then perform gaussian fits to each detected line to refine position and FWHM.

    Parameters
    ----------
    spectrum : The extracted arc spectrum to search for lines. It should be background-subtracted
        and must have an "uncertainty" attribute.

    fwhm : Estimated full-width half-maximum of the lines in pixels.

    window : The window size in units of fwhm to use for the gaussian fits.

    noise_factor : The factor to multiply the uncertainty by to determine the noise threshold
        in the `~specutils.fitting.find_lines_threshold` routine.

    Returns
    -------
    QTable
        A table of detected arc lines and their properties: centroid, fwhm, and amplitude.
    """
    # If fwhm is a float, convert it to a Quantity with the same unit as the spectral axis
    # of the input spectrum.
    if not isinstance(fwhm, u.Quantity):
        fwhm *= spectrum.spectral_axis.unit

    if fwhm.unit != spectrum.spectral_axis.unit:
        raise ValueError("fwhm must have the same units as spectrum.spectral_axis.")

    detected_lines = find_lines_threshold(spectrum, noise_factor=noise_factor)
    detected_lines = detected_lines[detected_lines['line_type'] == 'emission']

    centroids = []
    widths = []
    amplitudes = []
    for r in detected_lines:
        g_init = models.Gaussian1D(
            amplitude=spectrum.flux[r['line_center_index']],
            mean=r['line_center'],
            stddev=fwhm * gaussian_fwhm_to_sigma
        )
        g_fit = fit_lines(spectrum, g_init, window=window * fwhm)
        centroids.append(g_fit.mean.value * g_fit.mean.unit)
        widths.append(g_fit.stddev * gaussian_sigma_to_fwhm)
        amplitudes.append(g_fit.amplitude.value * g_fit.amplitude.unit)
    line_table = QTable()
    line_table['centroid'] = centroids
    line_table['fwhm'] = widths
    line_table['amplitude'] = amplitudes
    return line_table


def match_lines_wcs(
    pixel_positions: Sequence[float],
    catalog_wavelengths: Sequence[float],
    spectral_wcs: gWCS | astropy_WCS,
    tolerance: float = 5.0,
) -> QTable:
    """
    Use an input spectral WCS to match lines in an extracted spectrum to a catalog of known lines.
    Create matched table of pixel/wavelength positions for lines within a given tolerance of their
    WCS-predicted positions.

    Parameters
    ----------
    pixel_positions : The pixel positions of the lines in the calibration spectrum.

    catalog_wavelengths : The wavelengths of the lines in the catalog.

    spectral_wcs : The spectral WCS of the calibration spectrum.

    tolerance : The matching tolerance in pixels

    Returns
    -------
    QTable
        A table of the matched lines and their pixel/wavelength positions.
    """

    # This routine uses numpy broadcasting which doesn't always behave with Quantity objects.
    # Pull out the np.ndarray values to avoid those issues.
    if isinstance(pixel_positions, u.Quantity):
        pixel_positions = pixel_positions.value

    # Extra sanity handling to make sure the input Sequence can be converted to an np.array
    try:
        pixel_positions = np.array(pixel_positions, dtype=float)
    except ValueError as e:
        raise ValueError(f"pixel_positions must be convertable to np.array with dtype=float: {e}")

    catalog_pixels = spectral_wcs.world_to_pixel(catalog_wavelengths)
    separations = pixel_positions[:, np.newaxis] - catalog_pixels
    matched_loc = np.where(np.abs(separations) < tolerance)
    matched_table = QTable()
    matched_table["pixel_center"] = pixel_positions[matched_loc[0]] * u.pix
    matched_table["wavelength"] = catalog_wavelengths[matched_loc[1]]

    if len(matched_table) == 0:
        warnings.warn("No lines matched within the given tolerance.")

    return matched_table



import numpy as np
import pytest
from astropy import units as u
from astropy.modeling.fitting import LinearLSQFitter
from astropy.modeling.models import Polynomial1D
from astropy.table import QTable
from astropy.tests.helper import assert_quantity_allclose
from numpy.testing import assert_allclose

from specreduce import WavelengthCalibration1D


def test_linear_from_list(spec1d):
    centers = [0, 10, 20, 30]
    w = [5000, 5100, 5198, 5305]*u.AA
    test = WavelengthCalibration1D(spec1d, line_pixels=centers, line_wavelengths=w)
    spec2 = test.apply_to_spectrum(spec1d)

    assert_quantity_allclose(spec2.spectral_axis[0], 4998.8*u.AA)
    assert_quantity_allclose(spec2.spectral_axis[-1], 5495.169999*u.AA)


def test_wavelength_from_table(spec1d):
    centers = [0, 10, 20, 30]
    w = [5000, 5100, 5198, 5305]*u.AA
    table = QTable([w], names=["wavelength"])
    WavelengthCalibration1D(spec1d, line_pixels=centers, line_wavelengths=table)


def test_linear_from_table(spec1d):
    centers = [0, 10, 20, 30]
    w = [5000, 5100, 5198, 5305]*u.AA
    table = QTable([centers, w], names=["pixel_center", "wavelength"])
    test = WavelengthCalibration1D(spec1d, matched_line_list=table)
    spec2 = test.apply_to_spectrum(spec1d)

    assert_quantity_allclose(spec2.spectral_axis[0], 4998.8*u.AA)
    assert_quantity_allclose(spec2.spectral_axis[-1], 5495.169999*u.AA)


def test_poly_from_table(spec1d):
    # This test is mostly to prove that you can use other models
    centers = [0, 10, 20, 30, 40]
    w = [5005, 5110, 5214, 5330, 5438]*u.AA
    table = QTable([centers, w], names=["pixel_center", "wavelength"])

    test = WavelengthCalibration1D(spec1d, matched_line_list=table,
                                   input_model=Polynomial1D(2), fitter=LinearLSQFitter())
    test.apply_to_spectrum(spec1d)

    assert_allclose(test.fitted_model.parameters, [5.00477143e+03, 1.03457143e+01, 1.28571429e-02])


def test_replace_spectrum(spec1d, spec1d_with_emission_line):
    centers = [0, 10, 20, 30]*u.pix
    w = [5000, 5100, 5198, 5305]*u.AA
    test = WavelengthCalibration1D(spec1d, line_pixels=centers, line_wavelengths=w)
    # Accessing this property causes fits the model and caches the resulting WCS
    test.wcs
    assert "wcs" in test.__dict__

    # Replace the input spectrum, which should clear the cached properties
    test.input_spectrum = spec1d_with_emission_line
    assert "wcs" not in test.__dict__


def test_expected_errors(spec1d):
    centers = [0, 10, 20, 30, 40]
    w = [5005, 5110, 5214, 5330, 5438]*u.AA
    table = QTable([centers, w], names=["pixel_center", "wavelength"])

    with pytest.raises(ValueError, match="Cannot specify line_wavelengths separately"):
        WavelengthCalibration1D(spec1d, matched_line_list=table, line_wavelengths=w)

    with pytest.raises(ValueError, match="must have the same length"):
        w2 = [5005, 5110, 5214, 5330, 5438, 5500]*u.AA
        WavelengthCalibration1D(spec1d, line_pixels=centers, line_wavelengths=w2)

    with pytest.raises(ValueError, match="astropy.units.Quantity array or"
                                         " as an astropy.table.QTable"):
        w2 = [5005, 5110, 5214, 5330, 5438]
        WavelengthCalibration1D(spec1d, line_pixels=centers, line_wavelengths=w2)

    with pytest.raises(ValueError, match="specify at least one"):
        WavelengthCalibration1D(spec1d, line_pixels=centers)


def test_fit_residuals(spec1d):
    # test that fit residuals are all 0 when input is perfectly linear and model
    # is a linear model

    centers = np.array([0, 10, 20, 30])
    w = (0.5 * centers + 2) * u.AA
    test = WavelengthCalibration1D(spec1d, line_pixels=centers,
                                   line_wavelengths=w)

    test.apply_to_spectrum(spec1d)  # have to apply for residuals to be computed

    assert_quantity_allclose(test.residuals, 0.*u.AA, atol=1e-07*u.AA)


def test_fit_residuals_access(spec1d):
    # make sure that accessing residuals can be called before wcs/apply_to_spectrum

    centers = np.array([0, 10, 20, 30])
    w = (0.5 * centers + 2) * u.AA
    test = WavelengthCalibration1D(spec1d, line_pixels=centers,
                                   line_wavelengths=w)
    test.residuals
    test.wcs


def test_unsorted_pixels_wavelengths(spec1d):
    # make sure an error is raised if input matched pixels/wavelengths are
    # not strictly increasing or decreasing.

    centers = np.array([0, 10, 5, 30])
    w = (0.5 * centers + 2) * u.AA

    with pytest.raises(ValueError, match='Pixels must be strictly increasing or decreasing.'):
        WavelengthCalibration1D(spec1d, line_pixels=centers, line_wavelengths=w)

    # now test that it fails when wavelengths are unsorted
    centers = np.array([0, 10, 20, 30])
    w = np.array([2, 5, 6, 1]) * u.AA
    with pytest.raises(ValueError, match='Wavelengths must be strictly increasing or decreasing.'):
        WavelengthCalibration1D(spec1d, line_pixels=centers, line_wavelengths=w)

    # and same if those wavelengths are provided in a table
    table = QTable([w], names=["wavelength"])
    with pytest.raises(ValueError, match='Wavelengths must be strictly increasing or decreasing.'):
        WavelengthCalibration1D(spec1d, line_pixels=centers, line_wavelengths=table)

    # and again with decreasing pixels but unsorted wavelengths
    with pytest.raises(ValueError, match='Wavelengths must be strictly increasing or decreasing.'):
        WavelengthCalibration1D(spec1d, line_pixels=centers[::-1], line_wavelengths=w)



# Root level conftest.py is only for nice pytest header when using tox.

try:
    from pytest_astropy_header.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS
    ASTROPY_HEADER = True
except ImportError:
    ASTROPY_HEADER = False


def pytest_configure(config):

    if ASTROPY_HEADER:

        config.option.astropy_header = True

        # Customize the following lines to add/remove entries from the list of
        # packages for which version numbers are displayed when running the tests.
        PYTEST_HEADER_MODULES.pop('Pandas', None)
        PYTEST_HEADER_MODULES.pop('h5py', None)
        PYTEST_HEADER_MODULES['astropy'] = 'astropy'
        PYTEST_HEADER_MODULES['specutils'] = 'specutils'
        PYTEST_HEADER_MODULES['photutils'] = 'photutils'
        PYTEST_HEADER_MODULES['synphot'] = 'synphot'

        from specreduce import __version__
        TESTED_VERSIONS["specreduce"] = __version__



# This file is used to configure the behavior of pytest

import numpy as np
import pytest
from astropy import units as u
from astropy.io import fits
from astropy.nddata import CCDData, NDData, VarianceUncertainty
from astropy.utils.data import get_pkg_data_filename
from specutils import SpectralAxis

from specreduce.compat import SPECUTILS_LT_2, Spectrum

try:
    from pytest_astropy_header.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS
    ASTROPY_HEADER = True
except ImportError:
    ASTROPY_HEADER = False


# Test image is comprised of 30 rows with 10 columns each. Row content
# is row index itself. This makes it easy to predict what should be the
# value extracted from a region centered at any arbitrary Y position.
def _mk_test_data(imgtype, nrows=30, ncols=10):
    image_ones = np.ones(shape=(nrows, ncols))
    image = image_ones.copy()
    for j in range(nrows):
        image[j, ::] *= j
    if imgtype == "raw":
        pass  # no extra processing
    elif imgtype == "ccddata":
        image = CCDData(image, unit=u.Jy)
    else:  # spectrum
        flux = image * u.DN
        uncert = VarianceUncertainty(image_ones)
        if imgtype == "spec_no_axis":
            if SPECUTILS_LT_2:
                kwargs = {}
            else:
                kwargs = {"spectral_axis_index": image.ndim - 1}
            image = Spectrum(flux, uncertainty=uncert, **kwargs)
        else:  # "spec"
            image = Spectrum(flux, spectral_axis=np.arange(ncols) * u.um, uncertainty=uncert)
    return image


@pytest.fixture
def mk_test_img_raw():
    return _mk_test_data("raw")


@pytest.fixture
def mk_test_img():
    return _mk_test_data("ccddata")


@pytest.fixture
def mk_test_spec_no_spectral_axis():
    return _mk_test_data("spec_no_axis")


@pytest.fixture
def mk_test_spec_with_spectral_axis():
    return _mk_test_data("spec")


# Test data file already transposed like this:
# fn = download_file('https://stsci.box.com/shared/static/exnkul627fcuhy5akf2gswytud5tazmw.fits', cache=True)  # noqa: E501
# img = fits.getdata(fn).T
@pytest.fixture
def all_images():
    np.random.seed(7)

    filename = get_pkg_data_filename(
        "data/transposed_det_image_seq5_MIRIMAGE_P750Lexp1_s2d.fits", package="specreduce.tests")
    img = fits.getdata(filename)
    flux = img * (u.MJy / u.sr)
    sax = SpectralAxis(np.linspace(14.377, 3.677, flux.shape[-1]) * u.um)
    unc = VarianceUncertainty(np.random.rand(*flux.shape))

    if SPECUTILS_LT_2:
        kwargs = {}
    else:
        kwargs = {"spectral_axis_index": img.ndim - 1}

    all_images = {}
    all_images['arr'] = img
    all_images['s1d'] = Spectrum(flux, spectral_axis=sax, uncertainty=unc)
    all_images['s1d_pix'] = Spectrum(flux, uncertainty=unc, **kwargs)
    all_images['ccd'] = CCDData(img, uncertainty=unc, unit=flux.unit)
    all_images['ndd'] = NDData(img, uncertainty=unc, unit=flux.unit)
    all_images['qnt'] = img * flux.unit
    return all_images


@pytest.fixture
def spec1d():
    np.random.seed(7)
    flux = np.random.random(50)*u.Jy
    sa = np.arange(0, 50)*u.pix
    spec = Spectrum(flux, spectral_axis=sa)
    return spec


@pytest.fixture
def spec1d_with_emission_line():
    np.random.seed(7)
    sa = np.arange(0, 200)*u.pix
    flux = (np.random.randn(200) +
            10*np.exp(-0.01*((sa.value-130)**2)) +
            sa.value/100) * u.Jy
    spec = Spectrum(flux, spectral_axis=sa)
    return spec


@pytest.fixture
def spec1d_with_absorption_line():
    np.random.seed(7)
    sa = np.arange(0, 200)*u.pix
    flux = (np.random.randn(200) -
            10*np.exp(-0.01*((sa.value-130)**2)) +
            sa.value/100) * u.Jy
    spec = Spectrum(flux, spectral_axis=sa)
    return spec


def pytest_configure(config):

    if ASTROPY_HEADER:

        config.option.astropy_header = True

        # Customize the following lines to add/remove entries from the list of
        # packages for which version numbers are displayed when running the tests.
        PYTEST_HEADER_MODULES.pop('Pandas', None)
        PYTEST_HEADER_MODULES.pop('h5py', None)
        PYTEST_HEADER_MODULES['astropy'] = 'astropy'
        PYTEST_HEADER_MODULES['specutils'] = 'specutils'
        PYTEST_HEADER_MODULES['photutils'] = 'photutils'
        PYTEST_HEADER_MODULES['synphot'] = 'synphot'

        from specreduce import __version__
        TESTED_VERSIONS["specreduce"] = __version__



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import inspect
from copy import deepcopy
from dataclasses import dataclass
from typing import Literal

import numpy as np
from astropy import units as u
from astropy.nddata import VarianceUncertainty, NDData

from specreduce.compat import SPECUTILS_LT_2, Spectrum

__all__ = ["SpecreduceOperation"]

MaskingOption = Literal[
    "apply", "ignore", "propagate", "zero_fill", "nan_fill", "apply_mask_only", "apply_nan_only"
]

ImageLike = np.ndarray | NDData | u.Quantity


class _ImageParser:
    """
    Coerces images from accepted formats to Spectrum objects for
    internal use in specreduce's operation classes.

    Fills any and all of uncertainty, mask, units, and spectral axis
    that are missing in the provided image with generic values.
    Accepted image types are:

        - `~specutils.Spectrum1D` (preferred)
        - `~astropy.nddata.ccddata.CCDData`
        - `~astropy.nddata.ndddata.NDDData`
        - `~astropy.units.quantity.Quantity`
        - `~numpy.ndarray`
    """

    # The '_valid_mask_treatment_methods' in the Background, Trace, and Extract
    # classes is a subset of implemented methods.
    implemented_mask_treatment_methods = (
        "apply",
        "ignore",
        "propagate",
        "zero_fill",
        "nan_fill",
        "apply_mask_only",
        "apply_nan_only",
    )

    def _parse_image(
        self, image: ImageLike, disp_axis: int = 1, mask_treatment: MaskingOption = "apply"
    ) -> Spectrum:
        """
        Convert all accepted image types to a consistently formatted Spectrum object.

        Parameters
        ----------
        image : `~astropy.nddata.NDData`-like or array-like
            The image to be parsed. If None, defaults to class' own
            image attribute.
        disp_axis
            The index of the image's dispersion axis. Should not be
            changed until operations can handle variable image
            orientations.
        mask_treatment
            Specifies how to handle masked or non-finite values in the input image.
            The accepted values are:

            - ``apply``: The image remains unchanged, and any existing mask is combined\
                with a mask derived from non-finite values.
            - ``ignore``: The image remains unchanged, and any existing mask is dropped.
            - ``propagate``: The image remains unchanged, and any masked or non-finite pixel\
                causes the mask to extend across the entire cross-dispersion axis.
            - ``zero_fill``: Pixels that are either masked or non-finite are replaced with 0.0,\
                and the mask is dropped.
            - ``nan_fill``:  Pixels that are either masked or non-finite are replaced with nan,\
                and the mask is dropped.
            - ``apply_mask_only``: The  image and mask are left unmodified.
            - ``apply_nan_only``: The  image is left unmodified, the old mask is dropped, and a\
                new mask is created based on non-finite values.

        Returns
        -------
        Spectrum
        """
        # would be nice to handle (cross)disp_axis consistently across
        # operations (public attribute? private attribute? argument only?) so
        # it can be called from self instead of via kwargs...

        if image is None:
            # useful for Background's instance methods
            return self.image

        return self._get_data_from_image(image, disp_axis=disp_axis, mask_treatment=mask_treatment)

    @staticmethod
    def _get_data_from_image(
        image: ImageLike, disp_axis: int = 1, mask_treatment: MaskingOption = "apply"
    ) -> Spectrum:
        """
        Extract data array from various input types for `image`.

        Parameters
        ----------
        image
            Input image from which data is extracted. This can be a 2D numpy
            array, Quantity, or an NDData object.
        disp_axis
            The dispersion axis of the image.
        mask_treatment
            Specifies how to handle masked or non-finite values in the input image.

        Returns
        -------
        Spectrum
        """
        if isinstance(image, u.quantity.Quantity):
            img = image.value
        elif isinstance(image, np.ndarray):
            img = image
        else:  # NDData, including CCDData and Spectrum
            img = image.data

        mask = getattr(image, "mask", None)
        crossdisp_axis = (disp_axis + 1) % 2

        # next, handle masked and non-finite data in image.
        # A mask will be created from any non-finite image data, and combined
        # with any additional 'mask' passed in. If image is being parsed within
        # a specreduce operation that has 'mask_treatment' options, this will be
        # handled as well. Note that input data may be modified if a fill value
        # is chosen to handle masked data. The returned image will always have
        # `image.mask` even if there are no non-finite or masked values.
        img, mask = _ImageParser._mask_and_nonfinite_data_handling(
            image=img, mask=mask, mask_treatment=mask_treatment, crossdisp_axis=crossdisp_axis
        )

        # mask (handled above) and uncertainty are set as None when they aren't
        # specified upon creating a Spectrum object, so we must check whether
        # these attributes are absent *and* whether they are present but set as None
        if hasattr(image, "uncertainty"):
            uncertainty = image.uncertainty
        else:
            uncertainty = VarianceUncertainty(np.ones(img.shape))

        unit = getattr(image, "unit", u.Unit("DN"))

        spectral_axis = getattr(image, "spectral_axis", np.arange(img.shape[disp_axis]) * u.pix)

        if SPECUTILS_LT_2:
            kwargs = {}
        else:
            kwargs = {"spectral_axis_index": img.ndim - 1}
        img = Spectrum(
            img * unit, spectral_axis=spectral_axis, uncertainty=uncertainty, mask=mask,
            **kwargs
        )
        return img

    @staticmethod
    def _mask_and_nonfinite_data_handling(
        image: ImageLike,
        mask: ImageLike | None = None,
        mask_treatment: str = "apply",
        crossdisp_axis: int = 1,
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Handle the treatment of masked and non-finite data.

        All operations in Specreduce can take in a mask for the data as
        part of the input NDData.

        There are five options currently implemented for the treatment
        of masked and non-finite data - apply, ignore, zero_fill, nan_fill,
        apply_mask_only, and apply_nan_only. Depending on the routine,
        all or a subset of these three options are valid.

        Parameters
        ----------
        image : array-like
            The input image data array that may contain non-finite values.
        mask : array-like of bool or None
            An optional Boolean mask array. Non-finite values in the image will be added
            to this mask.
        mask_treatment
            Specifies how to handle masked or non-finite values in the input image.
        """
        if mask_treatment not in _ImageParser.implemented_mask_treatment_methods:
            raise ValueError(
                "'mask_treatment' must be one of "
                f"{_ImageParser.implemented_mask_treatment_methods}"
            )

        if mask is not None and (mask.dtype not in (bool, int)):
            raise ValueError("'mask' must be a boolean or integer array.")

        match mask_treatment:
            case "apply":
                mask = mask | (~np.isfinite(image)) if mask is not None else ~np.isfinite(image)
            case "ignore":
                mask = np.zeros(image.shape, dtype=bool)
            case "propagate":
                if mask is None:
                    mask = ~np.isfinite(image)
                else:
                    mask = mask | (~np.isfinite(image))
                mask[:] = mask.any(axis=crossdisp_axis, keepdims=True)
            case "zero_fill" | "nan_fill":
                mask = mask | (~np.isfinite(image)) if mask is not None else ~np.isfinite(image)
                image = deepcopy(image)
                if mask_treatment == "zero_fill":
                    image[mask] = 0.0
                else:
                    image[mask] = np.nan
                mask[:] = False
            case "apply_nan_only":
                mask = ~np.isfinite(image)
            case "apply_mask_only":
                mask = mask.copy() if mask is not None else np.zeros(image.shape, dtype=bool)

        if mask.all():
            raise ValueError("Image is fully masked. Check for invalid values.")

        return image, mask


@dataclass
class SpecreduceOperation(_ImageParser):
    """
    An operation to perform as part of a spectroscopic reduction pipeline.

    This class primarily exists to define the basic API for operations:
    parameters for the operation are provided at object creation,
    and then the operation object is called with the data objects required for
    the operation, which then *return* the data objects resulting from the
    operation.
    """

    def __call__(self):
        raise NotImplementedError("__call__ on a SpecreduceOperation needs to " "be overridden")

    @classmethod
    def as_function(cls, *args, **kwargs):
        """
        Run this operation as a function.  Syntactic sugar for e.g.,
        ``Operation.as_function(arg1, arg2, keyword=value)`` maps to
        ``Operation(arg2, keyword=value)(arg1)`` (if the ``__call__`` of
        ``Operation`` has only one argument)
        """
        argspec = inspect.getargs(cls.__call__.__code__)
        if argspec.varargs:
            raise NotImplementedError(
                "There is not a way to determine the "
                "number of inputs of a *args style "
                "operation"
            )
        ninputs = len(argspec.args) - 1

        callargs = args[:ninputs]
        noncallargs = args[ninputs:]
        op = cls(*noncallargs, **kwargs)
        return op(*callargs)



from copy import deepcopy

import pytest
import numpy as np

from astropy.units import DN
from astropy.nddata import NDData
from specreduce.core import _ImageParser


def mk_image():
    image = np.ones((3, 5))
    mask = np.zeros(image.shape, dtype=bool)
    image[1, 3] = np.nan
    mask[[0, 1], [1, 0]] = 1
    return NDData(image * DN, mask=mask)


def test_bad_option():
    image = mk_image()
    with pytest.raises(ValueError, match="'mask_treatment' must be one of"):
        _ImageParser()._parse_image(image, disp_axis=1, mask_treatment="bad")


def test_bad_mask_type():
    image = mk_image()
    image.mask = image.mask.astype(float)
    with pytest.raises(ValueError, match="'mask' must be a boolean or integer array."):
        _ImageParser()._parse_image(image, disp_axis=1, mask_treatment="apply")


def test_apply():
    image = mk_image()
    parsed_image = _ImageParser()._parse_image(deepcopy(image), disp_axis=1, mask_treatment="apply")
    np.testing.assert_array_equal(parsed_image.data, image.data)
    mask_true = np.zeros(image.data.shape, dtype=bool)
    mask_true[[0, 1, 1], [1, 0, 3]] = 1
    np.testing.assert_array_equal(parsed_image.mask, mask_true)

    image.mask = None
    parsed_image = _ImageParser()._parse_image(deepcopy(image), disp_axis=1, mask_treatment="apply")
    np.testing.assert_array_equal(parsed_image.data, image.data)
    mask_true = np.zeros(image.data.shape, dtype=bool)
    mask_true[1, 3] = 1
    np.testing.assert_array_equal(parsed_image.mask, mask_true)


def test_ignore():
    image = mk_image()
    parsed_image = _ImageParser()._parse_image(image, disp_axis=1, mask_treatment="ignore")
    np.testing.assert_array_equal(parsed_image.data, image.data)
    mask_true = np.zeros(image.data.shape, dtype=bool)
    np.testing.assert_array_equal(parsed_image.mask, mask_true)

    image.mask = None
    parsed_image = _ImageParser()._parse_image(image, disp_axis=1, mask_treatment="ignore")
    np.testing.assert_array_equal(parsed_image.data, image.data)
    mask_true = np.zeros(image.data.shape, dtype=bool)
    np.testing.assert_array_equal(parsed_image.mask, mask_true)


def test_propagate():
    image = mk_image()
    parsed_image = _ImageParser()._parse_image(image, disp_axis=1, mask_treatment="propagate")
    np.testing.assert_array_equal(parsed_image.data, image.data)
    mask_true = np.zeros(image.data.shape, dtype=bool)
    mask_true[:, [0, 1, 3]] = 1
    np.testing.assert_array_equal(parsed_image.mask, mask_true)

    image.mask = None
    parsed_image = _ImageParser()._parse_image(image, disp_axis=1, mask_treatment="propagate")
    np.testing.assert_array_equal(parsed_image.data, image.data)
    mask_true = np.zeros(image.data.shape, dtype=bool)
    mask_true[:, 3] = 1
    np.testing.assert_array_equal(parsed_image.mask, mask_true)


def test_zero_fill():
    image = mk_image()
    parsed_image = _ImageParser()._parse_image(image, disp_axis=1, mask_treatment="zero_fill")
    image_true = np.ones(image.data.shape)
    image_true[[0, 1, 1], [1, 0, 3]] = 0
    np.testing.assert_array_equal(parsed_image.data, image_true)
    mask_true = np.zeros(image.data.shape, dtype=bool)
    np.testing.assert_array_equal(parsed_image.mask, mask_true)

    image.mask = None
    parsed_image = _ImageParser()._parse_image(image, disp_axis=1, mask_treatment="zero_fill")
    image_true = np.ones(image.data.shape)
    image_true[1, 3] = 0
    np.testing.assert_array_equal(parsed_image.data, image_true)
    mask_true = np.zeros(image.data.shape, dtype=bool)
    np.testing.assert_array_equal(parsed_image.mask, mask_true)


def test_nan_fill():
    image = mk_image()
    parsed_image = _ImageParser()._parse_image(image, disp_axis=1, mask_treatment="nan_fill")
    image_true = np.ones(image.data.shape)
    image_true[[0, 1, 1], [1, 0, 3]] = np.nan
    np.testing.assert_array_equal(parsed_image.data, image_true)
    mask_true = np.zeros(image.data.shape, dtype=bool)
    np.testing.assert_array_equal(parsed_image.mask, mask_true)

    image.mask = None
    parsed_image = _ImageParser()._parse_image(image, disp_axis=1, mask_treatment="nan_fill")
    image_true = np.ones(image.data.shape)
    image_true[1, 3] = np.nan
    np.testing.assert_array_equal(parsed_image.data, image_true)
    mask_true = np.zeros(image.data.shape, dtype=bool)
    np.testing.assert_array_equal(parsed_image.mask, mask_true)


def test_apply_nan_only():
    image = mk_image()
    parsed_image = _ImageParser()._parse_image(image, disp_axis=1, mask_treatment="apply_nan_only")
    np.testing.assert_array_equal(parsed_image.data, image.data)
    mask_true = np.zeros(image.data.shape, dtype=bool)
    mask_true[1, 3] = 1
    np.testing.assert_array_equal(parsed_image.mask, mask_true)

    image.mask = None
    parsed_image = _ImageParser()._parse_image(image, disp_axis=1, mask_treatment="apply_nan_only")
    np.testing.assert_array_equal(parsed_image.data, image.data)
    mask_true = np.zeros(image.data.shape, dtype=bool)
    mask_true[1, 3] = 1
    np.testing.assert_array_equal(parsed_image.mask, mask_true)


def test_apply_mask_only():
    image = mk_image()
    parsed_image = _ImageParser()._parse_image(image, disp_axis=1, mask_treatment="apply_mask_only")
    np.testing.assert_array_equal(parsed_image.data, image.data)
    mask_true = np.zeros(image.data.shape, dtype=bool)
    mask_true[[0, 1], [1, 0]] = 1
    np.testing.assert_array_equal(parsed_image.mask, mask_true)

    image.mask = None
    parsed_image = _ImageParser()._parse_image(image, disp_axis=1, mask_treatment="apply_mask_only")
    np.testing.assert_array_equal(parsed_image.data, image.data)
    mask_true = np.zeros(image.data.shape, dtype=bool)
    np.testing.assert_array_equal(parsed_image.mask, mask_true)


def test_fully_masked():
    image = mk_image()
    image.mask[:] = 1
    with pytest.raises(ValueError, match="Image is fully masked."):
        _ImageParser()._parse_image(image, disp_axis=1, mask_treatment="apply")



import pytest

from specreduce.calibration_data import load_pypeit_calibration_lines


@pytest.mark.remote_data
def test_pypeit_single():
    """
    Test to load a single linelist from ``pypeit`` by passing a string.
    """
    line_tab = load_pypeit_calibration_lines('HeI', cache=True, show_progress=False)
    assert line_tab is not None
    assert "HeI" in line_tab['ion']
    assert sorted(list(line_tab.columns)) == [
        'Instr',
        'NIST',
        'Source',
        'amplitude',
        'ion',
        'wavelength'
    ]


@pytest.mark.remote_data
def test_pypeit_list():
    """
    Test to load and combine a set of linelists from ``pypeit`` by passing a list.
    """
    line_tab = load_pypeit_calibration_lines(['HeI', 'NeI'], cache=True, show_progress=False)
    assert line_tab is not None
    assert "HeI" in line_tab['ion']
    assert "NeI" in line_tab['ion']


@pytest.mark.remote_data
def test_pypeit_comma_list():
    """
    Test to load and combine a set of linelists from ``pypeit`` by passing a comma-separated list.
    """
    line_tab = load_pypeit_calibration_lines("HeI, NeI", cache=True, show_progress=False)
    assert line_tab is not None
    assert "HeI" in line_tab['ion']
    assert "NeI" in line_tab['ion']


@pytest.mark.remote_data
def test_pypeit_nonexisting_lamp():
    """
    Test to make sure a warning is raised if the lamp list includes a bad lamp name.
    """
    with pytest.warns(UserWarning, match='NeJ not in the list'):
        load_pypeit_calibration_lines(["HeI", "NeJ"], cache=True, show_progress=False)


@pytest.mark.remote_data
def test_pypeit_empty():
    """
    Test to make sure None is returned if an empty list is passed.
    """
    with pytest.warns(UserWarning, match='No calibration lines'):
        line_tab = load_pypeit_calibration_lines([], cache=True, show_progress=False)
    assert line_tab is None


@pytest.mark.remote_data
def test_pypeit_none():
    """
    Test to make sure None is returned if calibration lamp list is None.
    """
    line_tab = load_pypeit_calibration_lines(None, cache=True, show_progress=False)
    assert line_tab is None


@pytest.mark.remote_data
def test_pypeit_input_validation():
    """
    Check that bad inputs for ``pypeit`` linelists raise the appropriate warnings and exceptions
    """
    with pytest.raises(ValueError, match='.*Invalid calibration lamps specification.*'):
        _ = load_pypeit_calibration_lines({}, cache=True, show_progress=False)

    with pytest.warns() as record:
        _ = load_pypeit_calibration_lines(['HeI', 'ArIII'], cache=True, show_progress=False)
        if not record:
            pytest.fails("Expected warning about nonexistant linelist for ArIII.")



# specreduce/tests/test_wavecal1d.py
import astropy.units as u
import numpy as np
import pytest
from astropy.modeling import models
from astropy.nddata import StdDevUncertainty
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from numpy import array

from specreduce.compat import Spectrum
from specreduce.wavecal1d import WavelengthCalibration1D

ref_pixel = 250
pix_bounds = (0, 500)
p2w = models.Shift(-ref_pixel) | models.Polynomial1D(degree=3, c0=1, c1=0.2, c2=0.001)


@pytest.fixture
def mk_lines():
    obs_lines = ref_pixel + array([1, 2, 5, 8, 10])
    cat_lines = p2w(ref_pixel + array([1, 3, 5, 7, 8, 10]))
    return obs_lines, cat_lines


@pytest.fixture
def mk_matched_lines():
    obs_lines = ref_pixel + array([1, 2, 5, 8, 10])
    cat_lines = p2w(obs_lines)
    return obs_lines, cat_lines


@pytest.fixture
def mk_wc(mk_lines):
    obs_lines, cat_lines = mk_lines
    return WavelengthCalibration1D(
        obs_lines=obs_lines,
        line_lists=cat_lines,
        pix_bounds=(0, 10),
        ref_pixel=ref_pixel,
    )


@pytest.fixture
def mk_good_wc_with_transform(mk_lines):
    obs_lines, cat_lines = mk_lines
    wc = WavelengthCalibration1D(
        obs_lines=obs_lines,
        line_lists=cat_lines,
        pix_bounds=pix_bounds,
        ref_pixel=ref_pixel,
    )
    wc.solution.p2w = p2w
    return wc


@pytest.fixture
def mk_arc():
    return Spectrum(
        flux=np.ones(pix_bounds[1]) * u.DN,
        spectral_axis=np.arange(pix_bounds[1]) * u.pix,
        uncertainty=StdDevUncertainty(np.ones(pix_bounds[1])),
    )


def test_init(mk_arc, mk_lines):
    arc = mk_arc
    obs_lines, cat_lines = mk_lines
    WavelengthCalibration1D(arc_spectra=arc, line_lists=cat_lines, ref_pixel=ref_pixel)
    WavelengthCalibration1D(
        obs_lines=obs_lines,
        line_lists=cat_lines,
        pix_bounds=(0, 10),
        ref_pixel=ref_pixel,
    )

    with pytest.raises(ValueError, match="Only one of arc_spectra or obs_lines can be provided."):
        WavelengthCalibration1D(arc_spectra=arc, obs_lines=obs_lines, ref_pixel=ref_pixel)

    arc = Spectrum(
        flux=np.array([[1, 2, 3, 4, 5]]) * u.DN,
        spectral_axis=np.array([1, 2, 3, 4, 5]) * u.angstrom,
    )
    with pytest.raises(ValueError, match="The arc spectrum must be one dimensional."):
        WavelengthCalibration1D(arc_spectra=arc, ref_pixel=ref_pixel)

    with pytest.raises(ValueError, match="Must give pixel bounds when"):
        WavelengthCalibration1D(obs_lines=obs_lines, ref_pixel=ref_pixel)


def test_init_line_list(mk_arc):
    arc = mk_arc
    WavelengthCalibration1D(arc_spectra=arc, line_lists=["ArI"])
    WavelengthCalibration1D(arc_spectra=arc, line_lists="ArI")
    WavelengthCalibration1D(arc_spectra=arc, line_lists=[array([0.1])])
    WavelengthCalibration1D(arc_spectra=arc, line_lists=array([0.1]))
    WavelengthCalibration1D(arc_spectra=[arc, arc], line_lists=[["ArI"], ["ArI"]])
    WavelengthCalibration1D(arc_spectra=[arc, arc], line_lists=["ArI", ["ArI"]])
    WavelengthCalibration1D(arc_spectra=[arc, arc], line_lists=["ArI", "ArI"])
    WavelengthCalibration1D(arc_spectra=[arc, arc], line_lists=[array([0.1]), array([0.1])])
    WavelengthCalibration1D(arc_spectra=[arc, arc], line_lists=[array([0.1, 0.3]), ["ArI"]])
    with pytest.raises(ValueError, match="The number of line lists"):
        WavelengthCalibration1D(arc_spectra=[arc, arc], line_lists=[["ArI"]])


def test_find_lines(mocker, mk_arc):
    wc = WavelengthCalibration1D(arc_spectra=mk_arc)
    mock_find_arc_lines = mocker.patch("specreduce.wavecal1d.find_arc_lines")
    mock_find_arc_lines.return_value = {"centroid": np.array([5.0]) * u.angstrom}
    wc.find_lines(fwhm=2.0, noise_factor=1.5)
    assert wc._obs_lines is not None
    assert len(wc._obs_lines) == 1

    wc = WavelengthCalibration1D()
    with pytest.raises(ValueError, match="Must provide arc spectra to find lines."):
        wc.find_lines(fwhm=2.0, noise_factor=1.5)


def test_fit_lines(mk_matched_lines):
    lo, lc = mk_matched_lines
    wc = WavelengthCalibration1D(pix_bounds=pix_bounds, ref_pixel=ref_pixel)
    wc.fit_lines(pixels=lo, wavelengths=lc)
    assert wc.solution.p2w is not None
    assert wc.solution.p2w[1].degree == wc.degree

    wc = WavelengthCalibration1D(
        obs_lines=lo, line_lists=lc, pix_bounds=pix_bounds, ref_pixel=ref_pixel
    )
    wc.fit_lines(pixels=lo, wavelengths=lc, match_cat=True, match_obs=True)
    assert wc.solution.p2w is not None
    assert wc.solution.p2w[1].degree == wc.degree

    wc = WavelengthCalibration1D(pix_bounds=pix_bounds, ref_pixel=ref_pixel)
    with pytest.warns(UserWarning, match="The degree of the polynomial"):
        wc.fit_lines(degree=5, pixels=lo[:3], wavelengths=lc[:3])
    with pytest.raises(ValueError, match="Cannot fit without catalog"):
        wc.fit_lines(pixels=lo, wavelengths=lc, match_cat=True, match_obs=True)
    with pytest.raises(ValueError, match="Cannot fit without observed"):
        wc.fit_lines(pixels=lo, wavelengths=lc, match_cat=False, match_obs=True)


def test_observed_lines(mk_lines):
    wc = WavelengthCalibration1D()
    assert wc.observed_lines is None
    obs_lines, cat_lines = mk_lines
    wc = WavelengthCalibration1D(obs_lines=obs_lines, pix_bounds=pix_bounds, ref_pixel=ref_pixel)
    assert len(wc.observed_lines) == 1
    assert wc.observed_lines[0].shape == (len(obs_lines), 2)
    assert wc.observed_lines[0].mask.shape == (len(obs_lines), 2)
    np.testing.assert_allclose(wc.observed_lines[0].data[:, 0], obs_lines)
    assert np.all(wc.observed_lines[0].mask[:, 0] == 0)

    wc.observed_lines = obs_lines
    assert len(wc.observed_lines) == 1
    assert wc.observed_lines[0].mask.shape == (len(obs_lines), 2)
    np.testing.assert_allclose(wc.observed_lines[0].data[:, 0], obs_lines)
    assert np.all(wc.observed_lines[0].mask[:, 0] == 0)

    wc.observed_lines = wc.observed_lines
    assert len(wc.observed_lines) == 1
    assert wc.observed_lines[0].mask.shape == (len(obs_lines), 2)
    np.testing.assert_allclose(wc.observed_lines[0].data[:, 0], obs_lines)
    assert np.all(wc.observed_lines[0].mask[:, 0] == 0)

    # Line locations and amplitudes
    assert wc.observed_line_locations[0].shape == (len(obs_lines),)
    np.testing.assert_allclose(wc.observed_line_locations[0], obs_lines)
    assert wc.observed_line_amplitudes[0].shape == (len(obs_lines),)
    np.testing.assert_allclose(wc.observed_line_amplitudes[0], 0.0)


def test_catalog_lines(mk_lines):
    wc = WavelengthCalibration1D(ref_pixel=ref_pixel)
    assert wc.catalog_lines is None
    obs_lines, cat_lines = mk_lines
    wc = WavelengthCalibration1D(
        obs_lines=obs_lines, line_lists=cat_lines, pix_bounds=pix_bounds, ref_pixel=ref_pixel
    )
    assert len(wc.catalog_lines) == 1
    assert wc.catalog_lines[0].shape == (len(cat_lines), 2)
    np.testing.assert_allclose(wc.catalog_lines[0].data[:, 0], cat_lines)
    assert np.all(wc.catalog_lines[0].data[:, 1] == 0)

    wc.catalog_lines = cat_lines
    assert len(wc.catalog_lines) == 1
    np.testing.assert_allclose(wc.catalog_lines[0].data[:, 0], cat_lines)
    assert np.all(wc.catalog_lines[0].data[:, 1] == 0)

    wc.catalog_lines = wc.catalog_lines
    assert len(wc.catalog_lines) == 1
    np.testing.assert_allclose(wc.catalog_lines[0].data[:, 0], cat_lines)
    assert np.all(wc.catalog_lines[0].data[:, 1] == 0)


def test_fit_lines_raises_error_for_mismatched_sizes():
    pixels = array([2, 4, 6])
    wavelengths = p2w(array([2, 4, 5, 6]))
    wc = WavelengthCalibration1D(pix_bounds=pix_bounds, ref_pixel=ref_pixel)
    with pytest.raises(ValueError, match="The sizes of pixel and wavelength arrays must match."):
        wc.fit_lines(pixels=pixels, wavelengths=wavelengths)


def test_fit_lines_raises_error_for_insufficient_lines():
    pixels = [5]
    wavelengths = p2w(pixels)
    wc = WavelengthCalibration1D(pix_bounds=pix_bounds, ref_pixel=ref_pixel)
    with pytest.raises(ValueError, match="Need at least two lines for a fit"):
        wc.fit_lines(pixels=pixels, wavelengths=wavelengths)


def test_fit_lines_raises_error_for_missing_pixel_bounds():
    pixels = [2, 4, 6, 8]
    wavelengths = p2w(pixels)
    wc = WavelengthCalibration1D(ref_pixel=ref_pixel)
    with pytest.raises(ValueError, match="Cannot fit without pixel bounds set."):
        wc.fit_lines(pixels=pixels, wavelengths=wavelengths)


def test_fit_global():
    p2w = models.Shift(-ref_pixel) | models.Polynomial1D(degree=3, c0=650, c1=50.0, c2=-0.001)

    lines_obs = ref_pixel + array([2, 4, 4.5, 5, 6, 6.3, 8])
    lines_cat = p2w(ref_pixel + array([1, 2, 2.3, 4, 5, 6, 6.3, 7, 8, 9, 10]))
    wavelength_bounds = (649, 651)
    dispersion_bounds = (49, 51)
    wc = WavelengthCalibration1D(
        obs_lines=lines_obs, line_lists=lines_cat, pix_bounds=pix_bounds, ref_pixel=ref_pixel
    )
    ws = wc.fit_dispersion(wavelength_bounds, dispersion_bounds, popsize=10, refine_fit=True)
    np.testing.assert_allclose(ws.p2w[1].parameters, [650.0, 50.0, -0.001, 0.0], atol=1e-4)
    assert wc._fit is not None
    assert wc._fit.success
    assert wc.solution.p2w is not None

    wc = WavelengthCalibration1D(
        obs_lines=lines_obs, line_lists=lines_cat, pix_bounds=pix_bounds, ref_pixel=ref_pixel
    )
    wc.fit_dispersion(wavelength_bounds, dispersion_bounds, popsize=10, refine_fit=False)


def test_rms(mk_good_wc_with_transform):
    wc = mk_good_wc_with_transform
    assert np.isclose(wc.rms(space="wavelength"), 0)  # Perfect match, so RMS should be zero
    assert np.isclose(wc.rms(space="pixel"), 0)  # Perfect match, so RMS should be zero
    with pytest.raises(ValueError, match="Space must be either 'pixel' or 'wavelength'"):
        wc.rms(space="wavelenght")


def test_remove_unmatched_lines(mk_good_wc_with_transform):
    wc = mk_good_wc_with_transform
    wc.match_lines()
    wc.remove_unmatched_lines()
    assert wc.catalog_lines[0].size == wc.observed_lines[0].size


def test_plot_lines_with_valid_input():
    wc = WavelengthCalibration1D(ref_pixel=ref_pixel)
    wc.observed_lines = [np.ma.masked_array([100, 200, 300], mask=[False, True, False])]
    wc._cat_lines = wc.observed_lines
    fig = wc._plot_lines(kind="observed", frames=0, figsize=(8, 4), plot_labels=True)
    assert isinstance(fig, Figure)
    assert fig.axes[0].has_data()

    fig = wc._plot_lines(kind="catalog", frames=0, figsize=(8, 4), plot_labels=True)
    assert isinstance(fig, Figure)
    assert fig.axes[0].has_data()

    fig, ax = plt.subplots(1, 1)
    fig = wc._plot_lines(kind="catalog", frames=0, axes=ax, plot_labels=True)
    assert isinstance(fig, Figure)
    assert fig.axes[0].has_data()

    fig, axs = plt.subplots(1, 2)
    fig = wc._plot_lines(kind="catalog", frames=0, axes=axs, plot_labels=True)
    assert isinstance(fig, Figure)
    assert fig.axes[0].has_data()

    fig = wc._plot_lines(kind="observed", frames=0, axes=axs, plot_labels=True)
    assert isinstance(fig, Figure)
    assert fig.axes[0].has_data()


def test_plot_lines_raises_for_missing_transform(mk_wc):
    wc = mk_wc
    with pytest.raises(ValueError, match="Cannot map between pixels and"):
        wc._plot_lines(kind="observed", map_x=True)


def test_plot_lines_calls_transform_correctly(mk_good_wc_with_transform):
    wc = mk_good_wc_with_transform
    wc._plot_lines(kind="observed", map_x=True)
    wc._plot_lines(kind="catalog", map_x=True)


def test_plot_catalog_lines(mk_wc):
    wc = mk_wc
    wc.catalog_lines = [np.ma.masked_array([400, 500, 600], mask=[False, True, False])]
    fig = wc.plot_catalog_lines(frames=0, figsize=(10, 6), plot_labels=True, map_to_pix=False)
    assert isinstance(fig, Figure)
    assert fig.axes[0].has_data()

    fig, ax = plt.subplots(1, 1)
    fig = wc.plot_catalog_lines(frames=0, axes=ax, plot_labels=True)
    assert isinstance(fig, Figure)
    assert fig.axes[0].has_data()

    fig, axs = plt.subplots(1, 2)
    fig = wc.plot_catalog_lines(frames=[0], axes=axs, plot_labels=False)
    assert isinstance(fig, Figure)
    assert fig.axes[0].has_data()


def test_plot_observed_lines(mk_good_wc_with_transform, mk_arc):
    wc = mk_good_wc_with_transform
    wc.observed_lines = [np.ma.masked_array([100, 200, 300], mask=[False, True, False])]
    wc.arc_spectra = [mk_arc]
    for frames in [None, 0]:
        fig = wc.plot_observed_lines(frames=frames, figsize=(10, 5), plot_labels=True)
        assert isinstance(fig, Figure)
        assert fig.axes[0].has_data()
        assert len(fig.axes) == 1


def test_plot_fit(mk_arc, mk_good_wc_with_transform):
    wc = mk_good_wc_with_transform
    wc.arc_spectra = [mk_arc]
    for frames in [None, 0]:
        fig = wc.plot_fit(frames=frames, figsize=(12, 6), plot_labels=True)
        assert isinstance(fig, Figure)
        assert len(fig.axes) == 2
        assert fig.axes[0].has_data()
        assert fig.axes[1].has_data()
    wc.plot_fit(frames=frames, figsize=(12, 6), plot_labels=True, obs_to_wav=True)


def test_plot_residuals(mk_good_wc_with_transform):
    wc = mk_good_wc_with_transform

    fig = wc.plot_residuals(space="pixel", figsize=(8, 4))
    assert isinstance(fig, Figure)
    assert fig.axes[0].has_data()

    fig = wc.plot_residuals(space="wavelength", figsize=(8, 4))
    assert isinstance(fig, Figure)
    assert fig.axes[0].has_data()

    fig, ax = plt.subplots(1, 1)
    wc.plot_residuals(ax=ax, space="wavelength", figsize=(8, 4))

    with pytest.raises(ValueError, match="Invalid space specified"):
        wc.plot_residuals(space="wavelenght", figsize=(8, 4))



import numpy as np

from specreduce.core import _ImageParser
from specreduce.tracing import Trace, FlatTrace
from specreduce.extract import _ap_weight_image, _align_along_trace

__all__ = ['measure_cross_dispersion_profile', '_align_along_trace']


def measure_cross_dispersion_profile(image, trace=None, crossdisp_axis=0,
                                     width=None, pixel=None, pixel_range=None,
                                     statistic='median', align_along_trace=True):
    """
    Return a 1D (quantity) array of the cross-dispersion profile measured at a
    specified pixel value ('wavelength', but potentially before calibration),
    or the average profile across several pixel values (or range of pixel values)
    along the dispersion axis.

    If a single number is specified for `pixel`, then the profile at that pixel
    (i.e wavelength) will be returned. If several pixels are specified in a list or
    array, then they will be averaged (median or average, set by `statistic` which
    defaults to median). Alternatively, `pixel_range` can be specified as a tuple
    of integers specifying the minimum and maximum pixel range to average the
    profile. `pixel` and `pixel_range` cannot be set simultaneously, and the default
    is `pixel_range`=(min_image, max_image) to return an average profile across the
    entire wavelength range of the image.

    The window in the cross dispersion axis for measuring the profile at
    the pixel(s) specified is determined by `width` and `trace`. If a trace is
    provided (either as a Trace object, or as a number specifying the location
    on the cross-dispersion axis of a FlatTrace object) that will determine the
    center of the profile on the cross-dispersion axis. Otherwise, if `trace`
    is None, the center of the image will be the center of the returned profile
    (i.e., the center row assuming a horizontal trace).

    If `width` is none, a window size of half the image in the cross-dispersion
    axis will be used to measure the cross dispersion profile. Otherwise, an
    integer value can be set for `width` which will determine the size of the
    window around the trace used to measure the profile (this window moves with
    the trace if trace is not flat).

    By default, if a non-flat trace is used the image will be aligned along the
    trace. This can be controlled with the 'align_along_trace' parameter.

    Parameters
    ----------
    image : `~astropy.nddata.NDData`-like or array-like, required
        2D image to measure cross-dispersion profile.
    trace : Trace object, int, float, or None
        A Trace object, a number to specify a FlatTrace, or None to use the
        middle of the image. This is the position that defines the center of the
        cross-dispersion profile. [default: None]
    crossdisp_axis : int, optional
        The index of the image's cross-dispersion axis. [default: 0]
    width : tuple of int or None
        Width around 'trace' to calculate profile. If None, then all rows in the
        cross-dispersion axis will be used. [default: None]
    pixel: int, list of int, or None
        Pixel value(s) along the dispersion axis to return cross-dispersion profile.
        If several specified in list, then the average (method set by `statistic`)
        profile will be calculated. If None, and `pixel_range` is set, then
        `pixel_range` will be used. [default: None]
    pixel_range: tuple of int or None
        Tuple of (min, max) defining the pixel range (along dispersion axis) to
        calculate average cross-dispersion profile, up to and not inclusive of max.
        If None, and `pixel` is not None, `pixel` will be used. If None and `pixel`
        is also None, this will be interpreted as using the entire dispersion axis
        to generate an average profile for the whole image. [default: None]
    statistic: 'median' or 'average'
        If `pixel` specifies multiple pixels, or `pixel_range` is specified, an
        average profile will be returned. This can be either `median` (default)
        or `average` (mean). This is ignored if only one pixel is specified.
        [default: median]
    align_along_trace: bool
        Relevant only for non-flat traces. If True, "roll" each column so that
        the trace sits in the central row before calculating average profile. This
        will prevent any 'blurring' from averaging a non-flat trace at different
        pixel/wavelengths. [default: True]

    """

    if crossdisp_axis != 0 and crossdisp_axis != 1:
        raise ValueError('`crossdisp_axis` must be 0 or 1.')
    crossdisp_axis = int(crossdisp_axis)
    disp_axis = 1 if crossdisp_axis == 0 else 0

    unit = getattr(image, 'unit', None)

    # parse image, which will return a spectrum1D (note: this is not ideal,
    # but will be addressed at some point)
    parser = _ImageParser()
    image = parser._parse_image(image, disp_axis=disp_axis)

    # which we then need to make back into a masked array
    # again this way of parsing the image is not ideal but
    # thats just how it is for now.
    image = np.ma.MaskedArray(image.data, mask=image.mask)

    # transpose if disp_axis = 0 just for simplicity of calculations
    # image is already copied so this won't modify input
    if disp_axis == 0:
        image = image.T

    nrows = image.shape[crossdisp_axis]
    ncols = image.shape[disp_axis]

    if not isinstance(trace, Trace):  # `trace` can be a trace obj
        if trace is None:  # if None, make a FlatTrace in the center of image
            trace_pos = nrows / 2
            trace = FlatTrace(image, trace_pos)
        elif isinstance(trace, (float, int)):  # if float/int make a FlatTrace
            trace = FlatTrace(image, trace)
        else:
            raise ValueError('`trace` must be Trace object, number to specify '
                             'the location of a FlatTrace, or None to use center'
                             ' of image.')

    if statistic not in ['median', 'average']:
        raise ValueError("`statistic` must be 'median' or 'average'.")

    # determine if there is one pixel/wavelength selected or many as either a
    # list or a tuple to specify a range
    if pixel is not None:
        if pixel_range is not None:
            raise ValueError('Both `pixel` and `pixel_range` can not be set'
                             ' simultaneously.')
        if isinstance(pixel, (int, float)):
            pixels = np.array([int(pixel)])
        elif np.all([isinstance(x, (int, float)) for x in pixel]):
            pixels = np.array([int(x) for x in pixel])
        else:
            raise ValueError('`pixels` must be an integer, or list of integers '
                             'to specify where the crossdisperion profile should '
                             'be measured.')
    else:  # range is specified
        if pixel_range is None:
            pixels = np.arange(0, ncols)
        else:  # if not None, it should be a lower and upper bound
            if len(pixel_range) != 2:
                raise ValueError('`pixel_range` must be a tuple of integers.')
            pixels = np.arange(min(pixel_range), max(pixel_range))

    # now that we have all pixels that should be included in the profile, make
    # sure that they are within image bounds.
    # note: Should just warn instead and clip out out-of-bounds pixels, and only
    # warn if there are none left?
    if np.any(pixels < 0) or np.any(pixels > ncols - 1):
        raise ValueError('Pixels chosen to measure cross dispersion profile are'
                         ' out of image bounds.')

    # now that we know which pixel(s) on the disp. axis we want to include
    # figure out the range/window of pixels along the crossdisp axis to measure
    # the profile
    if width is None:  # if None, use all rows
        width = nrows
    elif isinstance(width, (float, int)):
        width = int(width)
    else:
        raise ValueError('`width` must be an integer, or None to use all '
                         'cross-dispersion pixels.')
        width = int(width)

    # rectify trace, if _align_along_trace is True and trace is not flat
    aligned_trace = None
    if align_along_trace:
        if not isinstance(trace, FlatTrace):
            # note: img was transposed according to `crossdisp_axis`: disp_axis will always be 1
            aligned_trace = _align_along_trace(image, trace.trace,
                                               disp_axis=1,
                                               crossdisp_axis=0)

            # new trace will be a flat trace at the center of the image
            trace_pos = nrows / 2
            trace = FlatTrace(aligned_trace, trace_pos)

    # create a weight image based on the trace and 'width' to mask around trace

    if width == nrows:
        wimg = np.zeros(image.shape)
    else:
        wimg = _ap_weight_image(trace, width, disp_axis, crossdisp_axis, image.shape)
        # invert mask to include, not exclude, pixels around trace
        wimg = (1 - wimg).astype(int)

    # now that we have figured out the mask for the window in cross-disp. axis,
    # select only the pixel(s) we want to include in measuring the avg. profile
    pixel_mask = np.ones((image.shape))
    pixel_mask[:, pixels] = 0

    # combine these masks to isolate the rows and cols used to measure profile
    combined_mask = np.logical_or(pixel_mask, wimg)

    if aligned_trace is not None:
        masked_arr = np.ma.MaskedArray(aligned_trace, combined_mask)
    else:
        masked_arr = np.ma.MaskedArray(image.data, combined_mask)

    # and measure the cross dispersion profile. if multiple pixels/wavelengths,
    # this will be an average. we already transposed data based on disp_axis so
    # axis is always 1 for this calculation
    if statistic == 'average':
        avg_prof = np.ma.mean(masked_arr, axis=1)
    else:  # must be median, we already checked.
        avg_prof = np.ma.median(masked_arr, axis=1)

    # and get profile
    avg_prof = avg_prof.data[~avg_prof.mask]

    # and re-apply original unit, if there was one
    if unit is not None:
        avg_prof *= unit

    return avg_prof



import numpy as np
import pytest
from astropy import units as u
from astropy.utils.exceptions import AstropyUserWarning

from specreduce.calibration_data import (
    AtmosphericExtinction,
    AtmosphericTransmission,
    SUPPORTED_EXTINCTION_MODELS
)


@pytest.mark.remote_data
def test_supported_models():
    """
    Test loading of supported models
    """
    for model in SUPPORTED_EXTINCTION_MODELS:
        ext = AtmosphericExtinction(model=model)
        assert len(ext.extinction_mag) > 0
        assert len(ext.transmission) > 0


@pytest.mark.remote_data
def test_custom_mag_model():
    """
    Test creation of custom model from Quantity arrays
    """
    wave = np.linspace(0.3, 2.0, 50)
    extinction = u.Magnitude(1. / wave, u.MagUnit(u.dimensionless_unscaled))
    ext = AtmosphericExtinction(extinction=extinction, spectral_axis=wave * u.um)
    assert len(ext.extinction_mag) > 0
    assert len(ext.transmission) > 0


@pytest.mark.remote_data
def test_custom_raw_mag_model():
    """
    Test creation of custom model from Quantity arrays
    """
    wave = np.linspace(0.3, 2.0, 50)
    extinction = 1. / wave * u.mag
    ext = AtmosphericExtinction(extinction=extinction, spectral_axis=wave * u.um)
    assert len(ext.extinction_mag) > 0
    assert len(ext.transmission) > 0


@pytest.mark.remote_data
def test_custom_linear_model():
    """
    Test creation of custom model from Quantity arrays
    """
    wave = np.linspace(0.3, 2.0, 50)
    extinction = 1. / wave * u.dimensionless_unscaled
    ext = AtmosphericExtinction(extinction=extinction, spectral_axis=wave * u.um)
    assert len(ext.extinction_mag) > 0
    assert len(ext.transmission) > 0


@pytest.mark.remote_data
def test_unsupported_model():
    """
    Test loading of a nonexistent model
    """
    with pytest.raises(ValueError, match='Requested extinction model,'):
        AtmosphericExtinction(model='bad_model')


@pytest.mark.remote_data
def test_missing_extinction_unit():
    """
    Test creation of custom model from Quantity arrays
    """
    wave = np.linspace(0.3, 2.0, 50)
    extinction = 1. / wave
    with pytest.warns(AstropyUserWarning):
        ext = AtmosphericExtinction(extinction=extinction, spectral_axis=wave * u.um)
    assert len(ext.extinction_mag) > 0
    assert len(ext.transmission) > 0


@pytest.mark.remote_data
def test_transmission_model():
    """
    Test creating of default atmospheric transmission model
    """
    ext = AtmosphericTransmission()
    assert len(ext.transmission) > 0

    with pytest.warns(RuntimeWarning):
        assert len(ext.extinction_mag) > 0



import pytest
from astropy.utils.exceptions import AstropyUserWarning

from specreduce.calibration_data import load_MAST_calspec, load_onedstds


@pytest.mark.remote_data
def test_load_MAST():
    sp = load_MAST_calspec("g191b2b_005.fits", show_progress=False)
    assert sp is not None
    assert len(sp.spectral_axis) > 0


@pytest.mark.remote_data
def test_load_MAST_bad_filename():
    with pytest.warns(AstropyUserWarning, match="Downloading of"):
        sp = load_MAST_calspec("j191b2b_005.fits", show_progress=False)
    assert sp is None


@pytest.mark.remote_data
def test_load_onedstds():
    sp = load_onedstds()
    assert sp is not None
    assert len(sp.spectral_axis) > 0


@pytest.mark.remote_data
def test_load_onedstds_bad_dataset():
    with pytest.warns(AstropyUserWarning, match="Specfied dataset,"):
        sp = load_onedstds("snffactory")
    assert sp is None


@pytest.mark.remote_data
def test_load_onedstds_bad_specfile():
    with pytest.warns(AstropyUserWarning, match="Can't load"):
        sp = load_onedstds(specfile="FG131.dat")
    assert sp is None



import astropy.units as u
import numpy as np
import pytest
from astropy.modeling import fitting, models
from astropy.nddata import NDData

from specreduce.compat import SPECUTILS_LT_2, Spectrum
from specreduce.tracing import FitTrace
from specreduce.utils.utils import measure_cross_dispersion_profile


def mk_gaussian_img(nrows=20, ncols=16, mean=10, stddev=4):
    """ Makes a simple horizontal gaussian image."""

    # note: this should become a fixture eventually, since other tests use
    # similar functions to generate test data.

    np.random.seed(7)
    col_model = models.Gaussian1D(amplitude=1, mean=mean, stddev=stddev)
    index_arr = np.tile(np.arange(nrows), (ncols, 1))

    return col_model(index_arr.T)


def mk_img_non_flat_trace(nrows=40, ncols=100, amp=10, stddev=2):
    """
    Makes an image with a gaussian source that has a non-flat trace dispersed
    along the x axis.
    """
    spec2d = np.zeros((nrows, ncols))

    for ii in range(spec2d.shape[1]):
        mgaus = models.Gaussian1D(amplitude=amp,
                                  mean=(9.+(20/spec2d.shape[1])*ii),
                                  stddev=stddev)
        rg = np.arange(0, spec2d.shape[0], 1)
        gaus = mgaus(rg)
        spec2d[:, ii] = gaus

    return spec2d


class TestMeasureCrossDispersionProfile():

    @pytest.mark.parametrize('pixel', [None, 1, [1, 2, 3]])
    @pytest.mark.parametrize('width', [10, 9])
    def test_measure_cross_dispersion_profile(self, pixel, width):
        """
        Basic test for `measure_cross_dispersion_profile`. Parametrized over
        different options for `pixel` to test using all wavelengths, a single
        wavelength, and a set of wavelengths, as well as different input types
        (plain array, quantity, Spectrum, and NDData), as well as `width` to
        use a window of all rows and a smaller window.
        """

        # test a few input formats
        images = []
        mean = 5.0
        stddev = 4.0
        dat = mk_gaussian_img(nrows=10, ncols=10, mean=mean, stddev=stddev)
        images.append(dat)  # test unitless
        images.append(dat * u.DN)
        images.append(NDData(dat * u.DN))
        if SPECUTILS_LT_2:
            kwargs = {}
        else:
            kwargs = {"spectral_axis_index": dat.ndim - 1}
        images.append(Spectrum(flux=dat * u.DN, **kwargs))

        for img in images:

            # use a flat trace at trace_pos=10, a window of width 10 around the trace
            # and use all 20 columns in image to create an average (median)
            # cross dispersion profile
            cdp = measure_cross_dispersion_profile(img, width=width, pixel=pixel)

            # make sure that if we fit a gaussian to the measured average profile,
            # that we get out the same profile that was used to create the image.
            # this should be exact since theres no noise in the image
            fitter = fitting.LevMarLSQFitter()
            mod = models.Gaussian1D()
            fit_model = fitter(mod, np.arange(width), cdp)

            assert fit_model.mean.value == np.where(cdp == max(cdp))[0][0]
            assert fit_model.stddev.value == stddev

            # test passing in a FlatTrace, and check the profile
            cdp = measure_cross_dispersion_profile(img, width=width, pixel=pixel)
            fit_model = fitter(mod, np.arange(width), cdp)
            assert fit_model.mean.value == np.where(cdp == max(cdp))[0][0]
            np.testing.assert_allclose(fit_model.stddev.value, stddev)

    @pytest.mark.filterwarnings("ignore:Model is linear in parameters")
    def test_cross_dispersion_profile_non_flat_trace(self):
        """
        Test measure_cross_dispersion_profile with a non-flat trace.
        Tests with 'align_along_trace' set to both True and False,
        to account for the changing center of the trace and measure
        the true profile shape, or to 'blur' the profile, respectivley.
        """

        image = mk_img_non_flat_trace()

        # fit the trace
        trace_fit = FitTrace(image)

        # when not aligning along trace and using the entire image
        # rows for the window, the center of the profile should follow
        # the shape of the trace
        peak_locs = [9, 10, 12, 13, 15, 16, 17, 19, 20, 22, 23, 24, 26, 27, 29]
        for i, pixel in enumerate(range(0, image.shape[1], 7)):
            profile = measure_cross_dispersion_profile(image,
                                                       trace=trace_fit,
                                                       width=None,
                                                       pixel=pixel,
                                                       align_along_trace=False,
                                                       statistic='average')
            peak_loc = (np.where(profile == max(profile))[0][0])
            assert peak_loc == peak_locs[i]

        # when align_along_trace = True, the shape of the profile should
        # not change since (there is some wiggling around though due to the
        # fact that the trace is rolled to the nearest integer value. this can
        # be smoothed with an interpolation option later on, but it is 'rough'
        # for now). In this test case, the peak positions will all either
        # be at pixel 20 or 21.
        for i, pixel in enumerate(range(0, image.shape[1], 7)):
            profile = measure_cross_dispersion_profile(image,
                                                       trace=trace_fit,
                                                       width=None,
                                                       pixel=pixel,
                                                       align_along_trace=True,
                                                       statistic='average')
            peak_loc = (np.where(profile == max(profile))[0][0])
            assert peak_loc in [20, 21]

    def test_errors_warnings(self):
        img = mk_gaussian_img(nrows=10, ncols=10)
        with pytest.raises(ValueError,
                           match='`crossdisp_axis` must be 0 or 1'):
            measure_cross_dispersion_profile(img, crossdisp_axis=2)

        with pytest.raises(ValueError, match='`trace` must be Trace object, '
                                             'number to specify the location '
                                             'of a FlatTrace, or None to use '
                                             'center of image.'):
            measure_cross_dispersion_profile(img, trace='not a trace or a number')

        with pytest.raises(ValueError, match="`statistic` must be 'median' "
                                             "or 'average'."):
            measure_cross_dispersion_profile(img, statistic='n/a')

        with pytest.raises(ValueError, match='Both `pixel` and `pixel_range` '
                                             'can not be set simultaneously.'):
            measure_cross_dispersion_profile(img, pixel=2, pixel_range=(2, 3))

        with pytest.raises(ValueError, match='`pixels` must be an integer, '
                                             'or list of integers to specify '
                                             'where the crossdisperion profile '
                                             'should be measured.'):
            measure_cross_dispersion_profile(img, pixel='str')

        with pytest.raises(ValueError, match='`pixel_range` must be a tuple '
                                             'of integers.'):
            measure_cross_dispersion_profile(img, pixel_range=(2, 3, 5))

        with pytest.raises(ValueError, match='Pixels chosen to measure cross '
                                             'dispersion profile are out of '
                                             'image bounds.'):
            measure_cross_dispersion_profile(img, pixel_range=(2, 12))

        with pytest.raises(ValueError, match='`width` must be an integer, '
                                             'or None to use all '
                                             'cross-dispersion pixels.'):
            measure_cross_dispersion_profile(img, width='.')
